# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import GraphTypesService_pb2 as GraphTypesService__pb2


class GraphTypesServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.GetVertexFrames = channel.unary_unary(
                '/xgtproto.GraphTypesService/GetVertexFrames',
                request_serializer=GraphTypesService__pb2.GetFramesMetadataRequest.SerializeToString,
                response_deserializer=GraphTypesService__pb2.GetFramesMetadataResponse.FromString,
                _registered_method=True)
        self.GetEdgeFrames = channel.unary_unary(
                '/xgtproto.GraphTypesService/GetEdgeFrames',
                request_serializer=GraphTypesService__pb2.GetFramesMetadataRequest.SerializeToString,
                response_deserializer=GraphTypesService__pb2.GetFramesMetadataResponse.FromString,
                _registered_method=True)
        self.GetTableFrames = channel.unary_unary(
                '/xgtproto.GraphTypesService/GetTableFrames',
                request_serializer=GraphTypesService__pb2.GetFramesMetadataRequest.SerializeToString,
                response_deserializer=GraphTypesService__pb2.GetFramesMetadataResponse.FromString,
                _registered_method=True)
        self.GetFrames = channel.unary_unary(
                '/xgtproto.GraphTypesService/GetFrames',
                request_serializer=GraphTypesService__pb2.GetFramesMetadataRequest.SerializeToString,
                response_deserializer=GraphTypesService__pb2.GetFramesMetadataResponse.FromString,
                _registered_method=True)
        self.CreateNamespace = channel.unary_unary(
                '/xgtproto.GraphTypesService/CreateNamespace',
                request_serializer=GraphTypesService__pb2.CreateNamespaceRequest.SerializeToString,
                response_deserializer=GraphTypesService__pb2.CreateNamespaceResponse.FromString,
                _registered_method=True)
        self.DeleteNamespace = channel.unary_unary(
                '/xgtproto.GraphTypesService/DeleteNamespace',
                request_serializer=GraphTypesService__pb2.DeleteNamespaceRequest.SerializeToString,
                response_deserializer=GraphTypesService__pb2.DeleteNamespaceResponse.FromString,
                _registered_method=True)
        self.GetNamespaces = channel.unary_unary(
                '/xgtproto.GraphTypesService/GetNamespaces',
                request_serializer=GraphTypesService__pb2.GetNamespacesRequest.SerializeToString,
                response_deserializer=GraphTypesService__pb2.GetNamespacesResponse.FromString,
                _registered_method=True)
        self.CreateFrame = channel.unary_unary(
                '/xgtproto.GraphTypesService/CreateFrame',
                request_serializer=GraphTypesService__pb2.CreateFrameRequest.SerializeToString,
                response_deserializer=GraphTypesService__pb2.GetFramesMetadataResponse.FromString,
                _registered_method=True)
        self.DeleteFrames = channel.unary_unary(
                '/xgtproto.GraphTypesService/DeleteFrames',
                request_serializer=GraphTypesService__pb2.DeleteFramesRequest.SerializeToString,
                response_deserializer=GraphTypesService__pb2.DeleteFramesResponse.FromString,
                _registered_method=True)
        self.ModifyColumns = channel.unary_unary(
                '/xgtproto.GraphTypesService/ModifyColumns',
                request_serializer=GraphTypesService__pb2.ModifyColumnsRequest.SerializeToString,
                response_deserializer=GraphTypesService__pb2.ModifyColumnsResponse.FromString,
                _registered_method=True)
        self.GetFrameSize = channel.unary_unary(
                '/xgtproto.GraphTypesService/GetFrameSize',
                request_serializer=GraphTypesService__pb2.GetFrameSizeRequest.SerializeToString,
                response_deserializer=GraphTypesService__pb2.GetFrameSizeResponse.FromString,
                _registered_method=True)
        self.GetFrameSchema = channel.unary_unary(
                '/xgtproto.GraphTypesService/GetFrameSchema',
                request_serializer=GraphTypesService__pb2.GetFrameSchemaRequest.SerializeToString,
                response_deserializer=GraphTypesService__pb2.GetFrameSchemaResponse.FromString,
                _registered_method=True)
        self.GetFrameLabels = channel.unary_unary(
                '/xgtproto.GraphTypesService/GetFrameLabels',
                request_serializer=GraphTypesService__pb2.GetFrameLabelsRequest.SerializeToString,
                response_deserializer=GraphTypesService__pb2.GetFrameLabelsResponse.FromString,
                _registered_method=True)
        self.GetRowLabelUniverse = channel.unary_unary(
                '/xgtproto.GraphTypesService/GetRowLabelUniverse',
                request_serializer=GraphTypesService__pb2.GetRowLabelUniverseRequest.SerializeToString,
                response_deserializer=GraphTypesService__pb2.GetRowLabelUniverseResponse.FromString,
                _registered_method=True)
        self.GetUserPermissions = channel.unary_unary(
                '/xgtproto.GraphTypesService/GetUserPermissions',
                request_serializer=GraphTypesService__pb2.GetUserPermissionsRequest.SerializeToString,
                response_deserializer=GraphTypesService__pb2.GetUserPermissionsResponse.FromString,
                _registered_method=True)
        self.GetUserLabels = channel.unary_unary(
                '/xgtproto.GraphTypesService/GetUserLabels',
                request_serializer=GraphTypesService__pb2.GetUserLabelsRequest.SerializeToString,
                response_deserializer=GraphTypesService__pb2.GetUserLabelsResponse.FromString,
                _registered_method=True)


class GraphTypesServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def GetVertexFrames(self, request, context):
        """Vertex frames.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetEdgeFrames(self, request, context):
        """Edge frames.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetTableFrames(self, request, context):
        """Table frames.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetFrames(self, request, context):
        """Any frames.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CreateNamespace(self, request, context):
        """Namespaces functions.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DeleteNamespace(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetNamespaces(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CreateFrame(self, request, context):
        """Multi-type functions.
        """
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DeleteFrames(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ModifyColumns(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetFrameSize(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetFrameSchema(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetFrameLabels(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetRowLabelUniverse(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetUserPermissions(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetUserLabels(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_GraphTypesServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'GetVertexFrames': grpc.unary_unary_rpc_method_handler(
                    servicer.GetVertexFrames,
                    request_deserializer=GraphTypesService__pb2.GetFramesMetadataRequest.FromString,
                    response_serializer=GraphTypesService__pb2.GetFramesMetadataResponse.SerializeToString,
            ),
            'GetEdgeFrames': grpc.unary_unary_rpc_method_handler(
                    servicer.GetEdgeFrames,
                    request_deserializer=GraphTypesService__pb2.GetFramesMetadataRequest.FromString,
                    response_serializer=GraphTypesService__pb2.GetFramesMetadataResponse.SerializeToString,
            ),
            'GetTableFrames': grpc.unary_unary_rpc_method_handler(
                    servicer.GetTableFrames,
                    request_deserializer=GraphTypesService__pb2.GetFramesMetadataRequest.FromString,
                    response_serializer=GraphTypesService__pb2.GetFramesMetadataResponse.SerializeToString,
            ),
            'GetFrames': grpc.unary_unary_rpc_method_handler(
                    servicer.GetFrames,
                    request_deserializer=GraphTypesService__pb2.GetFramesMetadataRequest.FromString,
                    response_serializer=GraphTypesService__pb2.GetFramesMetadataResponse.SerializeToString,
            ),
            'CreateNamespace': grpc.unary_unary_rpc_method_handler(
                    servicer.CreateNamespace,
                    request_deserializer=GraphTypesService__pb2.CreateNamespaceRequest.FromString,
                    response_serializer=GraphTypesService__pb2.CreateNamespaceResponse.SerializeToString,
            ),
            'DeleteNamespace': grpc.unary_unary_rpc_method_handler(
                    servicer.DeleteNamespace,
                    request_deserializer=GraphTypesService__pb2.DeleteNamespaceRequest.FromString,
                    response_serializer=GraphTypesService__pb2.DeleteNamespaceResponse.SerializeToString,
            ),
            'GetNamespaces': grpc.unary_unary_rpc_method_handler(
                    servicer.GetNamespaces,
                    request_deserializer=GraphTypesService__pb2.GetNamespacesRequest.FromString,
                    response_serializer=GraphTypesService__pb2.GetNamespacesResponse.SerializeToString,
            ),
            'CreateFrame': grpc.unary_unary_rpc_method_handler(
                    servicer.CreateFrame,
                    request_deserializer=GraphTypesService__pb2.CreateFrameRequest.FromString,
                    response_serializer=GraphTypesService__pb2.GetFramesMetadataResponse.SerializeToString,
            ),
            'DeleteFrames': grpc.unary_unary_rpc_method_handler(
                    servicer.DeleteFrames,
                    request_deserializer=GraphTypesService__pb2.DeleteFramesRequest.FromString,
                    response_serializer=GraphTypesService__pb2.DeleteFramesResponse.SerializeToString,
            ),
            'ModifyColumns': grpc.unary_unary_rpc_method_handler(
                    servicer.ModifyColumns,
                    request_deserializer=GraphTypesService__pb2.ModifyColumnsRequest.FromString,
                    response_serializer=GraphTypesService__pb2.ModifyColumnsResponse.SerializeToString,
            ),
            'GetFrameSize': grpc.unary_unary_rpc_method_handler(
                    servicer.GetFrameSize,
                    request_deserializer=GraphTypesService__pb2.GetFrameSizeRequest.FromString,
                    response_serializer=GraphTypesService__pb2.GetFrameSizeResponse.SerializeToString,
            ),
            'GetFrameSchema': grpc.unary_unary_rpc_method_handler(
                    servicer.GetFrameSchema,
                    request_deserializer=GraphTypesService__pb2.GetFrameSchemaRequest.FromString,
                    response_serializer=GraphTypesService__pb2.GetFrameSchemaResponse.SerializeToString,
            ),
            'GetFrameLabels': grpc.unary_unary_rpc_method_handler(
                    servicer.GetFrameLabels,
                    request_deserializer=GraphTypesService__pb2.GetFrameLabelsRequest.FromString,
                    response_serializer=GraphTypesService__pb2.GetFrameLabelsResponse.SerializeToString,
            ),
            'GetRowLabelUniverse': grpc.unary_unary_rpc_method_handler(
                    servicer.GetRowLabelUniverse,
                    request_deserializer=GraphTypesService__pb2.GetRowLabelUniverseRequest.FromString,
                    response_serializer=GraphTypesService__pb2.GetRowLabelUniverseResponse.SerializeToString,
            ),
            'GetUserPermissions': grpc.unary_unary_rpc_method_handler(
                    servicer.GetUserPermissions,
                    request_deserializer=GraphTypesService__pb2.GetUserPermissionsRequest.FromString,
                    response_serializer=GraphTypesService__pb2.GetUserPermissionsResponse.SerializeToString,
            ),
            'GetUserLabels': grpc.unary_unary_rpc_method_handler(
                    servicer.GetUserLabels,
                    request_deserializer=GraphTypesService__pb2.GetUserLabelsRequest.FromString,
                    response_serializer=GraphTypesService__pb2.GetUserLabelsResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'xgtproto.GraphTypesService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('xgtproto.GraphTypesService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class GraphTypesService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def GetVertexFrames(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/xgtproto.GraphTypesService/GetVertexFrames',
            GraphTypesService__pb2.GetFramesMetadataRequest.SerializeToString,
            GraphTypesService__pb2.GetFramesMetadataResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetEdgeFrames(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/xgtproto.GraphTypesService/GetEdgeFrames',
            GraphTypesService__pb2.GetFramesMetadataRequest.SerializeToString,
            GraphTypesService__pb2.GetFramesMetadataResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetTableFrames(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/xgtproto.GraphTypesService/GetTableFrames',
            GraphTypesService__pb2.GetFramesMetadataRequest.SerializeToString,
            GraphTypesService__pb2.GetFramesMetadataResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetFrames(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/xgtproto.GraphTypesService/GetFrames',
            GraphTypesService__pb2.GetFramesMetadataRequest.SerializeToString,
            GraphTypesService__pb2.GetFramesMetadataResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def CreateNamespace(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/xgtproto.GraphTypesService/CreateNamespace',
            GraphTypesService__pb2.CreateNamespaceRequest.SerializeToString,
            GraphTypesService__pb2.CreateNamespaceResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def DeleteNamespace(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/xgtproto.GraphTypesService/DeleteNamespace',
            GraphTypesService__pb2.DeleteNamespaceRequest.SerializeToString,
            GraphTypesService__pb2.DeleteNamespaceResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetNamespaces(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/xgtproto.GraphTypesService/GetNamespaces',
            GraphTypesService__pb2.GetNamespacesRequest.SerializeToString,
            GraphTypesService__pb2.GetNamespacesResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def CreateFrame(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/xgtproto.GraphTypesService/CreateFrame',
            GraphTypesService__pb2.CreateFrameRequest.SerializeToString,
            GraphTypesService__pb2.GetFramesMetadataResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def DeleteFrames(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/xgtproto.GraphTypesService/DeleteFrames',
            GraphTypesService__pb2.DeleteFramesRequest.SerializeToString,
            GraphTypesService__pb2.DeleteFramesResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ModifyColumns(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/xgtproto.GraphTypesService/ModifyColumns',
            GraphTypesService__pb2.ModifyColumnsRequest.SerializeToString,
            GraphTypesService__pb2.ModifyColumnsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetFrameSize(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/xgtproto.GraphTypesService/GetFrameSize',
            GraphTypesService__pb2.GetFrameSizeRequest.SerializeToString,
            GraphTypesService__pb2.GetFrameSizeResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetFrameSchema(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/xgtproto.GraphTypesService/GetFrameSchema',
            GraphTypesService__pb2.GetFrameSchemaRequest.SerializeToString,
            GraphTypesService__pb2.GetFrameSchemaResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetFrameLabels(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/xgtproto.GraphTypesService/GetFrameLabels',
            GraphTypesService__pb2.GetFrameLabelsRequest.SerializeToString,
            GraphTypesService__pb2.GetFrameLabelsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetRowLabelUniverse(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/xgtproto.GraphTypesService/GetRowLabelUniverse',
            GraphTypesService__pb2.GetRowLabelUniverseRequest.SerializeToString,
            GraphTypesService__pb2.GetRowLabelUniverseResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetUserPermissions(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/xgtproto.GraphTypesService/GetUserPermissions',
            GraphTypesService__pb2.GetUserPermissionsRequest.SerializeToString,
            GraphTypesService__pb2.GetUserPermissionsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetUserLabels(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/xgtproto.GraphTypesService/GetUserLabels',
            GraphTypesService__pb2.GetUserLabelsRequest.SerializeToString,
            GraphTypesService__pb2.GetUserLabelsResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



# -*- coding: utf-8 -*- ----------------------------------------------------===#
#
#  Copyright 2016-2025 Trovares Inc. dba Rocketgraph.  All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
#===------------------------------------------------------------------------===#

from __future__ import annotations

import grpc
import ipaddress
import logging
import operator
import os.path
import pyarrow
import sys
import warnings

from collections.abc import Iterable, Mapping, Sequence
from . import DataService_pb2 as data_proto

# In Python 3.9 and above, the abstract base classes (ABCs) from collections.abc
# can be used directly as type hints. This includes Iterable, Mapping, and
# Sequence, among others. For compatibility with versions below 3.9, we define
# aliases using the typing module, which has equivalent types with the same
# names. This ensures our code can use consistent type hints across different
# Python versions.
if sys.version_info >= (3, 9):
  # Directly use the ABCs from collections.abc in type hints since it is
  # supported. List and Dict type hints are directly available as built-in types
  # in Python 3.9+.
  Iter, Map, Seq, List, Dict = Iterable, Mapping, Sequence, list, dict
else:
   # For Python versions below 3.9, use the typing module for type hints.
  from typing import (Iterable as Iter, Mapping as Map, Sequence as Seq, List,
                                  Dict)

from typing import Optional, Union
from . import DataService_pb2 as data_proto
from . import ErrorMessages_pb2 as err_proto
from . import SchemaMessages_pb2 as sch_proto

log = logging.getLogger(__name__)

BOOLEAN = 'boolean'
INT = 'int'
UINT = 'uint'
FLOAT = 'float'
DATE = 'date'
TIME = 'time'
DATETIME = 'datetime'
IPADDRESS = 'ipaddress'
TEXT = 'text'
DURATION = 'duration'
ROWID = 'row_id'
LIST = 'list'
WGSPOINT = 'wgspoint'
CARTESIANPOINT = 'cartesianpoint'

# Send in 2MB chunks (grpc recommends 16-64 KB, but this got the best
# performance locally).  By default grpc only supports up to 4MB.
MAX_PACKET_SIZE = 2097152

# Default number of rows to use for Arrow data transfers between the client and
# the server.
DEFAULT_CHUNK_SIZE = 10000

class HeaderMode:
  NONE = 'none'
  IGNORE = 'ignore'
  NORMAL = 'normal'
  STRICT = 'strict'

  _all = [NONE,IGNORE,NORMAL,STRICT]

class XgtError(Exception):
  """
  Base exception class from which all other xgt exceptions inherit. It is
  raised in error cases that don't have a specific xgt exception type.
  """
  def __init__(self, msg, trace=''):
    self.msg = msg
    self.trace = trace

    if log.getEffectiveLevel() >= logging.DEBUG:
      if self.trace != '':
        log.debug(self.trace)
      else:
        log.debug(self.msg)
    Exception.__init__(self, self.msg)

class XgtNotImplemented(XgtError):
  """Raised for functionality with pending implementation."""

class XgtInternalError(XgtError):
  """
  Intended for internal server purposes only. This exception should not become
  visible to the user.
  """

class XgtIOError(XgtError):
  """An I/O problem occurred either on the client or server side."""
  def __init__(self, msg, trace='', job = None):
    self._job = job
    XgtError.__init__(self, msg, trace)

  @property
  def job(self):
    """
    Job: Job associated with the load/insert operation if available. May be
    None.
    """
    return self._job

class XgtServerMemoryError(XgtError):
  """
  The server memory usage is close to or at capacity and work could be lost.
  """

class XgtConnectionError(XgtError):
  """
  The client cannot properly connect to the server. This can include a failure
  to connect due to an xgt module version error.
  """

class XgtSyntaxError(XgtError):
  """A query was provided with incorrect syntax."""

class XgtTypeError(XgtError):
  """
  An unexpected type was supplied.

  For queries, an invalid data type was used either as an entity or as a
  property. For frames, either an edge, vertex or table frames was expected
  but the wrong frame type or some other data type was provided. For
  properties, the property declaration establishes the expected data type. A
  type error is raise if the data type used is not appropriate.
  """

class XgtValueError(XgtError):
  """An invalid or unexpected value was provided."""

class XgtNameError(XgtError):
  """
  An unexpected name was provided. Typically can occur during object retrieval
  where the object name was not found.
  """

class XgtArithmeticError(XgtError):
  """An invalid arithmetic calculation was detected and cannot be handled."""

class XgtFrameDependencyError(XgtError):
  """
  The requested action will produce an invalid graph or break a valid graph.
  """

class XgtTransactionError(XgtError):
  """A Transaction was attempted but didn't complete."""

class XgtSecurityError(XgtError):
  """A security violation occured."""

XgtErrorTypes = Union[XgtNotImplemented, XgtInternalError, XgtIOError,
                      XgtServerMemoryError, XgtConnectionError, XgtSyntaxError,
                      XgtTypeError, XgtValueError, XgtNameError,
                      XgtArithmeticError, XgtTransactionError, XgtSecurityError]

class _ContainerMap:
  """
  Maps container ids to frames.
  The user shouldn't manually construct this.

  Parameters
  ----------
  conn : Connection
    xgt Connection.
  container_dict : dictionary
    Dictionary containing keys of container ids to map.
    If empty is given, will map all frames.
  """
  def __init__(self, conn, container_dict = {}):
    self.frames = { }
    frames = conn.get_frames()

    for frame in frames:
      if not container_dict or frame._container_id in container_dict:
        self.frames[frame._container_id] = frame

  def get_frame(self, row_id):
    # The row_id is a 64 bit unsigned int where the 42 low bits are the
    # row position and the next higher 15 bits are the container ID.  The 7
    # high bits will never be set.
    container_id = row_id >> 42

    return self.frames[container_id] if container_id in self.frames else None

  def get_data(self, row_id, validation_id, include_row_labels = False):
    # The row_id is a 64 bit unsigned int where the 42 low bits are the
    # row position and the next higher 15 bits are the container ID.  The 7
    # high bits will never be set.
    container_id = row_id >> 42
    if container_id in self.frames:
      row_pos = row_id & 0x3FFFFFFFFFF
      frame = self.frames[container_id]

      return frame._get_data(frame.schema, offset = row_pos, length = 1,
                             include_row_labels = include_row_labels,
                             validation_id = validation_id,
                             absolute_indexing = True)[0]

    return None

class RowID:
  """
  RowID represents a row ID returned from a server.
  The user shouldn't manually construct these.

  Parameters
  ----------
  container_map : _ContainerMap
    Map of container ids to frames.
  row_id_list : list
    List in the form: [X, Y] where X is the row id and Y is the commit
    id when this value was valid.
  """
  def __init__(self, container_map, row_id_list):
    """
    Constructor for RowID. Should never be called directly by a user.
    """
    self._container_map = container_map
    self._row_id = row_id_list[0]
    self._validationid = row_id_list[1]

  def get_data(self, include_row_labels : bool = False) -> List:
    """
    Returns row data. If the frame this row points to has had deletions since
    the RowID was created, this row is considered invalid and will raise an
    exception.

    Parameters
    ----------
    include_row_labels : bool
      Indicates whether the security labels for each row should be egested
      along with the row.

    Returns
    -------
    list

    Raises
    ------
    XgtNameError
      If the frame does not exist on the server.
    XgtSecurityError
      If the user does not have required permissions for this action.
    XgtTransactionError
      If a conflict with another transaction occurs.
    ValueError
      If parameter is out of bounds or the row is no longer valid due to a
      transactional update.
    """
    return self._container_map.get_data(self._row_id, self._validationid,
                                        include_row_labels)

  @property
  def frame(self) -> Union[TableFrame, VertexFrame, EdgeFrame]:
    """
    TableFrame, VertexFrame, or EdgeFrame: The frame the RowID points into.
    """
    return self._container_map.get_frame(self._row_id)

  @property
  def position(self) -> int:
    """
    int: The row position the RowID points to.
    """
    # The row_id is a 64 bit unsigned int where the 42 low bits are the
    # row position and the next higher 15 bits are the container ID.  The 7
    # high bits will never be set.
    return self._row_id & 0x3FFFFFFFFFF

  @property
  def _validation_id(self):
    """
    int: The validation ID of the row id.
    """
    return self._validationid

  def __str__(self):
    return f"{{ ROW: {self.get_data()} }}"

  def __repr__(self):
    return str(self)

  def __eq__(self, obj):
    return self._row_id == obj._row_id and \
           self._validationid == obj._validationid

  def __lt__(self, obj):
    return self._row_id < obj._row_id

# Deprecation functions

def _deprecated(name, replacement, stacklevel = 0):
  # Display the caller of the function that calls _deprecated by default
  warnings.warn(f'{name} is deprecated and will be removed in future versions. '
                f'Use {replacement} instead.',
                DeprecationWarning, stacklevel = stacklevel + 3)

# Validation support functions

def _validated_schema_column(col):
  """
  Takes a user-supplied object representing a schema column and returns a
  valid schema column.

  Users can supply a variety of objects as valid schema columns. To simplify
  internal processing, we canonicalize these into a tuple containing name and
  type information, performing validation along the way.
  """
  if not isinstance(col, (list, tuple)):
    raise XgtTypeError('An input schema column must be of type tuple or list.')

  num_cols = len(col)
  if num_cols < 2 or num_cols > 4:
    raise XgtTypeError('An input schema column must contain between 2 and 4 '
                       'entries.')

  val_name = _validated_property_name(col[0])
  val_type = _validated_property_type(col[1])

  if val_type != "LIST":
    if num_cols > 2:
      raise XgtTypeError('A non-list input schema column must contain 2 '
                         'entries.')

    return (val_name, val_type)
  else:
    if num_cols < 3:
      raise XgtTypeError('A list input schema column must contain 3 or 4 '
                         'entries.')

    leaf_type = _validated_property_type(col[2])

    if num_cols == 3:
      return (val_name, val_type, leaf_type)
    else:
      if not isinstance(col[3], int):
        raise XgtTypeError('A list depth must be of integer type.')

      return (val_name, val_type, leaf_type, col[3])

def _validated_schema(obj, raise_on_empty_schema = True):
  """
  Takes a user-supplied object and returns a valid schema.

  Users can supply a variety of objects as valid schemas. To simplify internal
  processing, we canonicalize these into a list of tuples performing validation
  along the way. The inner tuples contain name and type information.
  """
  if not isinstance(obj, Iterable):
    raise XgtTypeError('The schema must be an Iterable with elements of '
                       'schema entries.')

  schema_returned = []
  for col in obj:
    schema_returned.append(_validated_schema_column(col))

  if raise_on_empty_schema and len(schema_returned) < 1:
    raise XgtTypeError('A schema cannot be empty.')

  return schema_returned

def _validated_columns(columns, schema):
  """
  Takes a user-supplied iterable of column indicators and returns a list of
  column positions sorted in the original schema order.

  The column indicators can be either column names from the schema or column
  positions.  Mixing is okay.
  """
  if columns is None:
    return None

  if not isinstance(columns, Iterable):
    raise XgtTypeError('columns must be an Iterable with elements of column '
                       'positions or column names.')

  fixed_columns = []
  schema_len = len(schema)
  for column in columns:
    if isinstance(column, int):
      if column < 0 or column >= schema_len:
        raise XgtValueError(f'Column position ({column}) out of bounds. '
                            f'Schema has {schema_len} columns.')

      if column not in fixed_columns:
        fixed_columns.append(column)
    elif isinstance(column, str):
      pos = -1
      for i, entry in enumerate(schema):
        if entry[0] == column:
          pos = i
          break

      if pos == -1:
        raise XgtValueError(f'Invalid column name: {column}.')

      if pos not in fixed_columns:
        fixed_columns.append(pos)
    else:
      raise XgtTypeError('Invalid column indicator. Must be column '
                         'positions or column names.')

  if isinstance(columns, set):
    fixed_columns = sorted(fixed_columns)

  return fixed_columns

def _validated_schema_columns(columns, schema):
  """
  Takes a user-supplied iterable of mixed column positions, column names, or
  schema entries and returns a valid schema.  Column positions and names must
  refer to valid columns in the schema.  Schema entries can refer to existing
  columns in the schema or new columns.
  """
  if columns is None:
    raise XgtTypeError('columns cannot be None.')

  if not isinstance(columns, Iterable):
    raise XgtTypeError('columns must be an Iterable with elements of column '
                       'positions, column names, or schema entries.')

  fixed_columns = []
  schema_len = len(schema)
  for column in columns:
    if isinstance(column, int):
      if column < 0 or column >= schema_len:
        raise XgtValueError(f'Column position ({column}) out of bounds. '
                            f'Schema has {schema_len} columns.')

      schema_col = schema[column]
    elif isinstance(column, str):
      schema_col = None
      for entry in schema:
        if entry[0] == column:
          schema_col = entry
          break

      if schema_col is None:
        raise XgtValueError(f'Invalid column name: {column}.')
    else:
      schema_col = column

    fixed_columns.append(_validated_schema_column(schema_col))

  if len(fixed_columns) < 1:
    raise XgtTypeError('A schema cannot be empty.')

  return fixed_columns

def _validated_frame_name(obj):
  """Takes a user-supplied object and returns a unicode frame name string."""
  _assert_isstring(obj)
  name = str(obj)
  if len(name) < 1:
    raise XgtNameError('Frame names cannot be empty.')
  return name

def _validated_namespace_name(obj):
  """Takes a user-supplied object and returns a unicode frame name string."""
  _assert_isstring(obj)
  name = str(obj)
  if len(name) < 1:
    raise XgtNameError('Namespace names cannot be empty.')
  return name

def _validated_property_name(obj):
  """Takes a user-supplied object and returns a unicode property name string."""
  _assert_isstring(obj)
  return str(obj)

def _get_valid_property_types_to_create():
  return [BOOLEAN, INT, UINT, FLOAT, DATE, TIME, DATETIME, IPADDRESS, TEXT,
          LIST, ROWID, DURATION, WGSPOINT, CARTESIANPOINT]

def _get_valid_property_types_for_return_only():
  return ['container_id', 'job_id']

def _validated_property_type(obj):
  """Takes a user-supplied object and returns an xGT schema type."""
  _assert_isstring(obj)
  prop_type = str(obj)
  valid_prop_types = _get_valid_property_types_to_create()
  if prop_type.lower() not in valid_prop_types:
    if prop_type.lower in _get_valid_property_types_for_return_only():
      raise XgtTypeError(f'Invalid schema property type "{prop_type}". This '
                         'type cannot be used when creating a frame.')
    else:
      raise XgtTypeError(f'Invalid schema property type "{prop_type}"')
  return prop_type.upper()

def _validate_opt_level(optlevel):
  """
  Valid optimization level values are:
    - 0: No optimization.
    - 1: General optimization.
    - 2: WHERE-clause optimization.
    - 3: Degree-cycle optimization.
    - 4: Query order optimization.
  """
  if isinstance(optlevel, int):
    if optlevel not in [0, 1, 2, 3, 4]:
      raise XgtValueError(f"Invalid optlevel '{optlevel}'")
  else:
    raise XgtTypeError("optlevel must be an integer")
  return True

def _assert_noerrors(response):
  if len(response.error) > 0:
    error = response.error[0]
    try:
      error_code_name = err_proto.ErrorCodeEnum.Name(error.code)
      error_class = _code_error_map[error_code_name]
      raise error_class(error.message, error.detail)
    except XgtError:
      raise
    except Exception as ex:
      raise XgtError(f"Error detected while raising exception{str(ex)}",
                     str(ex))

def _convert_flight_server_error_into_xgt(error):
  if len(error.extra_info) >= 8 and error.extra_info[0:6] == b"ERROR:":
    try:
      error_class = _code_error_map[
          err_proto.ErrorCodeEnum.Name(int(error.extra_info[6:8]))]
      return error_class(str(error), error.extra_info)
    except:
      pass
  return XgtError(str(error))

def _assert_isstring(value):
  if not isinstance(value, str):
    raise TypeError(f"{str(value)} is not a string")

_code_error_map = {
  'GENERIC_ERROR': XgtError,
  'NOT_IMPLEMENTED': XgtNotImplemented,
  'INTERNAL_ERROR': XgtInternalError,
  'IO_ERROR': XgtIOError,
  'SERVER_MEMORY_ERROR': XgtServerMemoryError,
  'CONNECTION_ERROR': XgtConnectionError,
  'SYNTAX_ERROR': XgtSyntaxError,
  'TYPE_ERROR': XgtTypeError,
  'VALUE_ERROR': XgtValueError,
  'NAME_ERROR': XgtNameError,
  'ARITHMETIC_ERROR': XgtArithmeticError,
  'FRAME_DEPENDENCY_ERROR': XgtFrameDependencyError,
  'TRANSACTION_ERROR': XgtTransactionError,
  'SECURITY_ERROR': XgtSecurityError,
}

def _verify_offset_length(offset, length):
  max_uint64 = sys.maxsize * 2 + 1

  if not isinstance(offset, (int, str)):
    raise XgtTypeError("offset must be an integer.")

  if isinstance(offset, str):
    offset = int(offset)

  if offset < 0:
    raise XgtValueError("offset must be a non-negative integer.")
  if offset > max_uint64:
    raise XgtValueError(f"offset must be < {max_uint64}")

  if length is not None:
    if not isinstance(length, (int, str)):
      raise XgtTypeError("length must be an integer.")

    if isinstance(length, str):
      length = int(length)

    if length < 0:
      raise XgtValueError("length must be a non-negative integer.")
    if length > max_uint64:
      raise XgtValueError(f"length must be < {max_uint64}")

  return offset, length

def _verify_row_positions(rows, frame_or_job):
  if rows is None:
    return

  if not isinstance(rows, Iterable):
    raise XgtTypeError('rows must be an Iterable with elements of integer row '
                       'positions.')

  # Get the number of rows.  Handle that a job could return None.
  num_rows = frame_or_job.num_rows
  if num_rows is None:
    num_rows = 0

  for row in rows:
    if not isinstance(row, int):
      raise XgtTypeError('Invalid row indicator. Must be integer row position.')

    if row < 0 or row >= num_rows:
      raise XgtValueError(f'Row position {row} out of bounds. Must be in '
                          f'range 0 to {num_rows - 1}.')

def _create_flight_ticket(frame_or_job, name, offset, length, rows = None,
                          columns = None, include_row_labels = False,
                          row_label_column_header = None,
                          order = True, date_as_string = False,
                          job_id = None, validation_id = None,
                          duration_as_interval = False,
                          row_filter = None, absolute_indexing = False,
                          **kwargs):
  if rows is not None and (offset != 0 or length is not None):
    raise XgtValueError('Cannot give rows when giving either offset or length.')

  offset, length = _verify_offset_length(offset, length)
  _verify_row_positions(rows, frame_or_job)

  ticket = f'`{name}`'

  if offset != 0:
    ticket += f".offset={offset}"
  if length is not None:
    ticket += f".length={length}"
  if rows is not None:
    ticket += ".rows=" + ','.join([str(x) for x in rows])
  if columns is not None:
    ticket += ".columns=" + ','.join([str(x) for x in columns])
  if order:
    ticket += ".order=True"
  if date_as_string:
    ticket += ".dates_as_strings=True"
  if include_row_labels:
    ticket += ".egest_row_labels=True"
  if row_label_column_header is not None:
    ticket += ".label_column_header=" + row_label_column_header
  if validation_id is not None:
    ticket += f".validation_id={str(validation_id)}"
  if duration_as_interval:
    ticket += ".duration_as_interval=True"
  if absolute_indexing:
    ticket += ".absolute_indexing=True"

  if row_filter is not None:
    row_filter_value = f'.row_filter="{row_filter}"'
    ticket += row_filter_value

  for k,v in kwargs.items():
    if k == "record_history":
      ticket += f".record_history={str(v)}"
    else:
      raise ValueError(f"kwarg {k} not supported.")
  if len(kwargs) == 0:
    # For get_data
    ticket += ".record_history=false"

  if job_id is not None:
    if isinstance(job_id, str):
      job_id = int(job_id)
    elif not isinstance(job_id, int):
      raise ValueError("job ID must be an int.")
    ticket += f".job_id={job_id}"

  return ticket

def _convert_row_id_list(container_map, value, depth = 0):
  if value is None:
    return None
  elif depth > 1:
    return [_convert_row_id_list(container_map, list_val, depth - 1)
            if list_val != None else None for list_val in value]
  else:
    return [RowID(container_map, id_val) if id_val != None else None
            for id_val in value]

def _convert_row_id_list_pandas(container_map, value, depth = 0):
  import numpy as np

  if value is None:
    return None

  # Pyarrow's converions function to_pandas() converts nested pyarrow lists to
  # nested 1D numpy arrays, not a single multi-dimensional numpy array. To get
  # a numpy array of separate numpy arrays, the array has to be created with
  # object type and the correct size first and other numpy arrays copied into
  # the positions.  Otherwise, a multidimensional numpy array will be created.
  v = np.ndarray(shape=(len(value),), dtype = 'object')

  if depth > 1:
    for i, entry in enumerate(value):
      v[i] = _convert_row_id_list_pandas(container_map, entry, depth - 1) \
             if entry is not None else None
  else:
    for i, entry in enumerate(value):
      v[i] = RowID(container_map, entry) if entry is not None else None

  return v

def _convert_ip_address_list(value, depth = 1):
  if value is None:
    return None
  elif depth > 1:
    return [_convert_ip_address_list(list_val, depth - 1)
            if list_val is not None else None for list_val in value]
  else:
    return [ipaddress.ip_address(id_val) if id_val is not None else None
            for id_val in value]

def _convert_ip_address_list_pandas(value, depth = 1):
  import numpy as np

  if value is None:
    return None

  # Pyarrow's converions function to_pandas() converts nested pyarrow lists to
  # nested 1D numpy arrays, not a single multi-dimensional numpy array. To get
  # a numpy array of separate numpy arrays, the array has to be created with
  # object type and the correct size first and other numpy arrays copied into
  # the positions.  Otherwise, a multidimensional numpy array will be created.
  v = np.ndarray(shape=(len(value),), dtype = 'object')

  if depth > 1:
    for i, entry in enumerate(value):
      v[i] = _convert_ip_address_list_pandas(entry, depth - 1) \
             if entry is not None else None
  else:
    for i, entry in enumerate(value):
      v[i] = ipaddress.ip_address(entry) if entry is not None else None

  return v

# Creates a list of conversion functions from the schema.
# Currently has conversions for row id and paths(row id lists).
def _schema_row_conversion(schema, conn):
  container_map = None
  for entry in schema:
    if entry[1] == ROWID or (entry[1] == LIST and entry[2] == ROWID):
      container_map = _ContainerMap(conn)
      break

  conversion_funcs = []
  for entry in schema:
    if entry[1] == IPADDRESS:
      def closure(value):
        return ipaddress.ip_address(value) if value is not None else None
      conversion_funcs.append(closure)
    elif(entry[1] == LIST and entry[2] == IPADDRESS):
      if len(entry) == 3:
        def closure(value):
          return _convert_ip_address_list(value)
        conversion_funcs.append(closure)
      else:
        length = entry[3]
        def closure(value):
          return _convert_ip_address_list(value, length)
        conversion_funcs.append(closure)
    elif entry[1] == ROWID:
      def closure(value):
        return RowID(container_map, value) if value != None else None
      conversion_funcs.append(closure)
    elif(entry[1] == LIST and entry[2] == ROWID):
      if len(entry) == 3:
        def closure(value):
          return _convert_row_id_list(container_map, value, 1)
        conversion_funcs.append(closure)
      else:
        length = entry[3]
        def closure(value):
          return _convert_row_id_list(container_map, value, length)
        conversion_funcs.append(closure)
    else:
      conversion_funcs.append(None)

  return conversion_funcs

def _get_data_python_from_table(arrow_table, schema, conn):
  if arrow_table is None:
    return None

  # Get functions used to convert columns coming from the server.
  # Currently, this provides functions for converting paths(lists of row ids)
  # and row ids into the RowID class.
  conversion_funcs = _schema_row_conversion(schema, conn)

  # List comprehension here is simpler, but has slow performance due to the
  # access pattern being bad for the cache hits.
  return_list = [None] * arrow_table.num_rows
  if arrow_table.num_columns > len(conversion_funcs):
    conversion_funcs += [None] * (arrow_table.num_columns -
                                  len(conversion_funcs))

  for i in range(arrow_table.num_rows):
    return_list[i] = []
  for i, x in enumerate(arrow_table):
    if conversion_funcs[i] == None:
      for j, y in enumerate(x):
        return_list[j].append(y.as_py())
    else:
      for j, y in enumerate(x):
        return_list[j].append(conversion_funcs[i](y.as_py()))

  return return_list

def _get_data_pandas_from_table(arrow_table, schema, conn):
  import pandas as pd

  if arrow_table is None:
    return None

  pandas_dict = {}
  container_map = None
  for entry in schema:
    if entry[1] == ROWID or (entry[1] == LIST and entry[2] == ROWID):
      container_map = _ContainerMap(conn)
      break
  schema_len = len(schema)
  for i, (name, col) in enumerate(zip(arrow_table.column_names, arrow_table)):
    if i < schema_len and schema[i][1] == IPADDRESS:
      pandas_dict[name] = \
          pd.Series([None if x.as_py() is None else ipaddress.ip_address(x)
                     for x in col], dtype = 'object')
    elif i < schema_len and schema[i][1] == ROWID:
      pandas_dict[name] = \
          pd.Series([None if x.as_py() is None else RowID(container_map,
                                                          x.as_py())
                     for x in col], dtype = 'object')
    elif i < schema_len and schema[i][1] == LIST and schema[i][2] == IPADDRESS:
      depth = 1 if len(schema[i]) == 3 else schema[i][3]
      pandas_dict[name] = pd.Series([_convert_ip_address_list_pandas(x.as_py(),
                                                                     depth)
                                     for x in col])
    elif i < schema_len and schema[i][1] == LIST and schema[i][2] == ROWID:
      depth = 1 if len(schema[i]) == 3 else schema[i][3]
      pandas_dict[name] = \
          pd.Series([None if x.as_py() is None else
                     _convert_row_id_list_pandas(container_map, x.as_py(),
                                                 depth)
                     for x in col], dtype = 'object')
    else:
      pandas_dict[name] = col.to_pandas()

  return pd.DataFrame(pandas_dict)

def _get_data_arrow(conn, ticket):
  try:
    res_table = conn.arrow_conn.do_get(pyarrow.flight.Ticket(ticket)).read_all()
    return res_table
  except pyarrow._flight.FlightServerError as err:
    raise _convert_flight_server_error_into_xgt(err) from err
  except pyarrow._flight.FlightUnavailableError as err:
    raise XgtConnectionError(str(err)) from err

def _get_data_python(conn, ticket, schema):
  res_table = _get_data_arrow(conn, ticket)
  return _get_data_python_from_table(res_table, schema, conn)

def _get_data_pandas(conn, ticket, schema):
  res_table = _get_data_arrow(conn, ticket)
  return _get_data_pandas_from_table(res_table, schema, conn)

# Helper functions for low code.

# Convert the pyarrow type to an xgt type.
def _pyarrow_type_to_xgt_type(pyarrow_type, depth = 0):
  if pyarrow.types.is_boolean(pyarrow_type):
    return (BOOLEAN, depth)
  elif (pyarrow.types.is_timestamp(pyarrow_type) or
        pyarrow.types.is_date64(pyarrow_type)):
    return (DATETIME, depth)
  elif pyarrow.types.is_date(pyarrow_type):
    return (DATE, depth)
  elif pyarrow.types.is_time(pyarrow_type):
    return (TIME, depth)
  elif pyarrow.types.is_temporal(pyarrow_type):
    return (DURATION, depth)
  elif pyarrow.types.is_signed_integer(pyarrow_type):
    return (INT, depth)
  elif pyarrow.types.is_unsigned_integer(pyarrow_type):
    return (UINT, depth)
  elif (pyarrow.types.is_float32(pyarrow_type) or
        pyarrow.types.is_float64(pyarrow_type) or
        pyarrow.types.is_decimal(pyarrow_type)):
    return (FLOAT, depth)
  elif (pyarrow.types.is_string(pyarrow_type) or
        pyarrow.types.is_large_string(pyarrow_type)):
    return (TEXT, depth)
  elif (pyarrow.types.is_list(pyarrow_type) or
        pyarrow.types.is_large_list(pyarrow_type)):
    return _pyarrow_type_to_xgt_type(pyarrow_type.value_type, depth + 1)
  else:
    raise XgtTypeError(
        f"Cannot convert pyarrow type {str(pyarrow_type)} to xGT type.")

def _infer_xgt_schema_from_pyarrow_schema(pyarrow_schema):
  xgt_schema = []
  for c in pyarrow_schema:
    xgt_type = _pyarrow_type_to_xgt_type(c.type)
    if xgt_type[1] == 0:
      xgt_schema.append([c.name, xgt_type[0]])
    else:
      xgt_schema.append([c.name, LIST, xgt_type[0], xgt_type[1]])
  return xgt_schema

# Get the column in the schema by name or position.
def _find_key_in_schema(key, schema):
  if isinstance(key, str):
    found_key = False
    for elem in schema:
      if elem[0] == key:
        found_key = True
        break
    if not found_key:
      raise XgtNameError(
          f"The key {str(key)} not found in schema: {str(schema)}")
    return key
  elif isinstance(key, int):
    if key >= len(schema) or key < 0:
      raise XgtError(f"Could not locate key {str(key)} in schema with "
                     f"{len(schema)} entries.")
    return schema[key][0]

# Modify an xgt schema based on a frame column name to data column name
# mapping. The key names of this map will become the column names of the
# new schema. The values of the map correspond to the columns of the
# initial schema.
def _apply_mapping_to_schema(initial_schema, frame_to_data_column_mapping):
  def find_data_col_name(data_col):
    if isinstance(data_col, str):
      return data_col
    elif isinstance(data_col, int):
      if data_col >= len(initial_schema) or data_col < 0:
        err = ("Error creating the schema. The column mapping refers to "
               f"data column position {data_col}, but only "
               f"{len(initial_schema)} columns were found in the data.")
        raise XgtValueError(err)

      return initial_schema[data_col][0]

  data_col_name_to_type = { elem[0] : elem[1] for elem in initial_schema }

  schema = []
  for frame_col, data_col in frame_to_data_column_mapping.items():
    data_type = data_col_name_to_type[find_data_col_name(data_col)]
    schema.append([frame_col, data_type])

  return schema

def _remove_label_columns_from_schema(initial_schema, row_label_columns):
  def find_data_col_name(data_col):
    if isinstance(data_col, str):
      return data_col
    elif isinstance(data_col, int):
      if data_col >= len(initial_schema) or data_col < 0:
        err = ("Error creating the schema. The row_label_columns parameter "
               f"refers to data column position {data_col}, but only "
               f"{len(initial_schema)} columns were found in the data.")
        raise XgtValueError(err)
      return initial_schema[data_col][0]

  data_col_name_to_type = { elem[0] : elem[1] for elem in initial_schema }

  label_columns = set([find_data_col_name(col) for col in row_label_columns])

  return [col for col in initial_schema if col[0] not in label_columns]

def _generate_proto_schema_from_xgt_schema(request, schema):
  for col in schema:
    prop = sch_proto.Property()
    prop.name = col[0]
    prop.data_type = sch_proto.UvalTypeEnum.Value(col[1])

    if (prop.data_type == sch_proto.UvalTypeEnum.Value('LIST')):
      prop.leaf_type = sch_proto.UvalTypeEnum.Value(col[2])

      if (len(col) > 3):
        if col[3] > 1:
          prop.list_depth = col[3] - 1
        elif col[3] == 1:
          prop.list_depth = 0
        else:
          raise XgtTypeError("A list cannot have a depth less than one.")
      else:
        prop.list_depth = 0

    request.schema.property.extend([prop])

# Compares two dot-separated number versions like "2.10.8".
# Note: This function does not support versions with alpha/beta characters
# or any non-integer segments and will return None for such inputs.
# An example failure case would be: "2.10.8a"
def _compare_versions(version1, version2, op=operator.eq):
  try:
    v1 = list(map(int, version1.split('.')))
    v2 = list(map(int, version2.split('.')))
    return op(v1, v2)
  except:
    return None

def _validate_column_mapping_in_ingest(column_mapping):
  error_msg = ('The data type of "column_mapping" is incorrect. '
               'Expects a dictionary with string keys and string '
               'or integer values.')
  if column_mapping is not None:
    if not isinstance(column_mapping, Mapping):
      raise TypeError(error_msg)
    for frame_col, file_col in column_mapping.items():
      if not isinstance(file_col, (str, int)):
        raise TypeError(error_msg)
      if not isinstance(frame_col, str):
        raise TypeError(error_msg)

def _set_column_mapping_in_ingest_request(column_mapping):
  if column_mapping is not None:
    return_value = data_proto.ColumnMapping()
    return_value.has_mapping = True
    for frame_col, file_col in column_mapping.items():
      if isinstance(file_col, str):
        name_mapping = data_proto.FrameColumnToSourceName()
        name_mapping.frame_column_name = _validated_property_name(frame_col)
        name_mapping.file_column_name = file_col
        return_value.frame_column_to_source_name.extend(
            [name_mapping])
      elif isinstance(file_col, int):
        idx_mapping = data_proto.FrameColumnToSourceIdx()
        idx_mapping.frame_column_name = _validated_property_name(frame_col)
        idx_mapping.file_column_idx = file_col
        return_value.frame_column_to_source_idx.extend([idx_mapping])
      else:
        raise TypeError('The data type of "column_mapping" is incorrect. '
                        'Expects a dictionary with string keys and string '
                        'or integer values.')
  return return_value

def _validate_client_path(one_path):
  if one_path.endswith('.gz') or one_path.endswith('.bz2'):
    raise XgtNotImplemented(f'Loading/Saving compressed files from a local '
                            f'filesystem is not supported: {one_path}')

def _group_paths(paths, for_ingest):
  client_prefix = 'xgt://'
  server_prefix = 'xgtd://'
  url_prefixes = ['s3://', 'https://', 'http://', 'ftps://', 'ftp://']
  client_paths = []
  server_paths = []
  url_paths = []
  if isinstance(paths, str):
    paths = [paths]
  elif not isinstance(paths, (list, tuple)):
    return client_paths, server_paths, url_paths
  for one_path in paths:
    one_path_normalized = one_path.lower()
    if one_path == "":
      raise ValueError('one or more "paths" are empty')
    if one_path_normalized.startswith(client_prefix):
      _validate_client_path(one_path)
      client_paths.append(one_path[len(client_prefix):])
    elif one_path_normalized.startswith(server_prefix):
      server_paths.append(one_path[len(server_prefix):])
    elif any(map(lambda p: one_path_normalized.startswith(p), url_prefixes)):
      for url_prefix in url_prefixes:
        if one_path_normalized.startswith(url_prefix):
          url_paths.append(one_path)
          break
    else:
      if '://' in one_path:
        msg = f'Unsupported url protocol in path "{one_path}".'
        raise XgtNotImplemented(msg)
      _validate_client_path(one_path)
      client_paths.append(one_path)
  return client_paths, server_paths, url_paths

def _split_local_paths(paths):
  parquet_paths = []
  other_paths = []
  for path in paths:
    extension = os.path.splitext(path)[-1]
    if extension == ".parquet":
      parquet_paths.append(path)
    else:
      other_paths.append(path)

  return parquet_paths, other_paths

def _convert_header_mode(header_mode, request):
  if header_mode == HeaderMode.IGNORE:
    request.header_mode = data_proto.IGNORE_HEADERS
  elif header_mode == HeaderMode.NORMAL:
    request.header_mode = data_proto.NORMAL
  elif header_mode == HeaderMode.STRICT:
    request.header_mode = data_proto.STRICT
  else:
    request.header_mode = data_proto.NONE



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: SchemaMessages.proto
# Protobuf Python Version: 5.28.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    28,
    1,
    '',
    'SchemaMessages.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import ErrorMessages_pb2 as ErrorMessages__pb2
from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x14SchemaMessages.proto\x12\x08xgtproto\x1a\x13\x45rrorMessages.proto\x1a\x1fgoogle/protobuf/timestamp.proto\"\x82\x01\n\x08Property\x12\x0c\n\x04name\x18\x01 \x01(\t\x12)\n\tdata_type\x18\x02 \x01(\x0e\x32\x16.xgtproto.UvalTypeEnum\x12)\n\tleaf_type\x18\x03 \x01(\x0e\x32\x16.xgtproto.UvalTypeEnum\x12\x12\n\nlist_depth\x18\x04 \x01(\x04\".\n\x06Schema\x12$\n\x08property\x18\x01 \x03(\x0b\x32\x12.xgtproto.Property\"\x87\x02\n\x08\x41nyFrame\x12\x0c\n\x04name\x18\x01 \x01(\t\x12 \n\x06schema\x18\x02 \x01(\x0b\x32\x10.xgtproto.Schema\x12%\n\x04type\x18\x03 \x01(\x0e\x32\x17.xgtproto.FrameTypeEnum\x12\x12\n\nvertex_key\x18\x04 \x01(\t\x12\x12\n\nsource_key\x18\x05 \x01(\t\x12\x12\n\ntarget_key\x18\x06 \x01(\t\x12\x15\n\rsource_vertex\x18\x07 \x01(\t\x12\x15\n\rtarget_vertex\x18\x08 \x01(\t\x12\x14\n\x0c\x63ontainer_id\x18\t \x01(\x04\x12\x11\n\tcommit_id\x18\n \x01(\x04\x12\x11\n\tdelete_id\x18\x0b \x01(\x04\"\xc8\x01\n\x10QueryEdgeMessage\x12\x34\n\x0e\x65\x64ge_direction\x18\x01 \x01(\x0e\x32\x1c.xgtproto.QueryEdgeDirection\x12\x12\n\nedge_label\x18\x02 \x01(\t\x12\x12\n\nedge_frame\x18\x03 \x01(\t\x12\x14\n\x0csource_label\x18\x04 \x01(\t\x12\x14\n\x0csource_frame\x18\x05 \x01(\t\x12\x14\n\x0ctarget_label\x18\x06 \x01(\t\x12\x14\n\x0ctarget_frame\x18\x07 \x01(\t\"b\n\rQueryEdgeList\x12)\n\x05\x65\x64ges\x18\x01 \x03(\x0b\x32\x1a.xgtproto.QueryEdgeMessage\x12\x17\n\nplan_score\x18\x02 \x01(\x01H\x00\x88\x01\x01\x42\r\n\x0b_plan_score\"?\n\tQueryPlan\x12\x32\n\x11plan_as_edge_list\x18\x01 \x03(\x0b\x32\x17.xgtproto.QueryEdgeList\"\xaa\x04\n\tJobStatus\x12\x1e\n\x05\x65rror\x18\x01 \x03(\x0b\x32\x0f.xgtproto.Error\x12\x0e\n\x06job_id\x18\x02 \x01(\x04\x12\'\n\x06status\x18\x03 \x01(\x0e\x32\x17.xgtproto.JobStatusEnum\x12.\n\nstart_time\x18\x04 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12,\n\x08\x65nd_time\x18\x05 \x01(\x0b\x32\x1a.google.protobuf.Timestamp\x12\x0c\n\x04user\x18\x06 \x01(\t\x12\x13\n\x0b\x64\x65scription\x18\x07 \x01(\t\x12<\n\rvisited_edges\x18\x08 \x03(\x0b\x32%.xgtproto.JobStatus.VisitedEdgesEntry\x12\x0e\n\x06timing\x18\t \x03(\t\x12 \n\x06schema\x18\n \x01(\x0b\x32\x10.xgtproto.Schema\x12\x10\n\x08num_rows\x18\x0b \x01(\x04\x12\'\n\nquery_plan\x18\x0c \x01(\x0b\x32\x13.xgtproto.QueryPlan\x12\x1b\n\x13total_ingest_errors\x18\r \x01(\x04\x12\x14\n\x0cingest_error\x18\x0e \x03(\t\x12\x19\n\x11\x64\x65\x66\x61ult_namespace\x18\x0f \x01(\t\x12\x15\n\rresults_frame\x18\x10 \x01(\t\x1a\x33\n\x11VisitedEdgesEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\r\n\x05value\x18\x02 \x01(\x04:\x02\x38\x01*\xfc\x01\n\x0cUvalTypeEnum\x12\x15\n\x11UNKNOWN_UVAL_TYPE\x10\x00\x12\x0b\n\x07UNKNOWN\x10\x01\x12\x0b\n\x07\x42OOLEAN\x10\x02\x12\x07\n\x03INT\x10\x03\x12\t\n\x05\x46LOAT\x10\x04\x12\x08\n\x04\x44\x41TE\x10\x05\x12\x08\n\x04TIME\x10\x06\x12\x0c\n\x08\x44\x41TETIME\x10\x07\x12\r\n\tIPADDRESS\x10\x08\x12\x08\n\x04TEXT\x10\t\x12\x10\n\x0c\x43ONTAINER_ID\x10\n\x12\n\n\x06JOB_ID\x10\x0b\x12\x08\n\x04LIST\x10\x0c\x12\n\n\x06ROW_ID\x10\r\x12\x0c\n\x08\x44URATION\x10\x0e\x12\x08\n\x04UINT\x10\x0f\x12\x0c\n\x08WGSPOINT\x10\x10\x12\x12\n\x0e\x43\x41RTESIANPOINT\x10\x11*A\n\rFrameTypeEnum\x12\t\n\x05TABLE\x10\x00\x12\n\n\x06VERTEX\x10\x01\x12\x08\n\x04\x45\x44GE\x10\x02\x12\x0f\n\x0b\x41SSOCIATIVE\x10\x03*z\n\rJobStatusEnum\x12\x16\n\x12UNKNOWN_JOB_STATUS\x10\x00\x12\r\n\tSCHEDULED\x10\x01\x12\x0b\n\x07RUNNING\x10\x02\x12\r\n\tCOMPLETED\x10\x03\x12\x0c\n\x08\x43\x41NCELED\x10\x04\x12\n\n\x06\x46\x41ILED\x10\x05\x12\x0c\n\x08ROLLBACK\x10\x06*=\n\x12QueryEdgeDirection\x12\x0b\n\x07\x46ORWARD\x10\x00\x12\x0c\n\x08\x42\x41\x43KWARD\x10\x01\x12\x0c\n\x08NOT_EDGE\x10\x02\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'SchemaMessages_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_JOBSTATUS_VISITEDEDGESENTRY']._loaded_options = None
  _globals['_JOBSTATUS_VISITEDEDGESENTRY']._serialized_options = b'8\001'
  _globals['_UVALTYPEENUM']._serialized_start=1461
  _globals['_UVALTYPEENUM']._serialized_end=1713
  _globals['_FRAMETYPEENUM']._serialized_start=1715
  _globals['_FRAMETYPEENUM']._serialized_end=1780
  _globals['_JOBSTATUSENUM']._serialized_start=1782
  _globals['_JOBSTATUSENUM']._serialized_end=1904
  _globals['_QUERYEDGEDIRECTION']._serialized_start=1906
  _globals['_QUERYEDGEDIRECTION']._serialized_end=1967
  _globals['_PROPERTY']._serialized_start=89
  _globals['_PROPERTY']._serialized_end=219
  _globals['_SCHEMA']._serialized_start=221
  _globals['_SCHEMA']._serialized_end=267
  _globals['_ANYFRAME']._serialized_start=270
  _globals['_ANYFRAME']._serialized_end=533
  _globals['_QUERYEDGEMESSAGE']._serialized_start=536
  _globals['_QUERYEDGEMESSAGE']._serialized_end=736
  _globals['_QUERYEDGELIST']._serialized_start=738
  _globals['_QUERYEDGELIST']._serialized_end=836
  _globals['_QUERYPLAN']._serialized_start=838
  _globals['_QUERYPLAN']._serialized_end=901
  _globals['_JOBSTATUS']._serialized_start=904
  _globals['_JOBSTATUS']._serialized_end=1458
  _globals['_JOBSTATUS_VISITEDEDGESENTRY']._serialized_start=1407
  _globals['_JOBSTATUS_VISITEDEDGESENTRY']._serialized_end=1458
# @@protoc_insertion_point(module_scope)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import AdminService_pb2 as AdminService__pb2


class AdminServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.ChangeErrorCount = channel.unary_unary(
                '/xgtproto.AdminService/ChangeErrorCount',
                request_serializer=AdminService__pb2.ChangeErrorCountRequest.SerializeToString,
                response_deserializer=AdminService__pb2.ChangeErrorCountResponse.FromString,
                _registered_method=True)
        self.FreeUserMemorySize = channel.unary_unary(
                '/xgtproto.AdminService/FreeUserMemorySize',
                request_serializer=AdminService__pb2.FreeUserMemorySizeRequest.SerializeToString,
                response_deserializer=AdminService__pb2.FreeUserMemorySizeResponse.FromString,
                _registered_method=True)
        self.MaxUserMemorySize = channel.unary_unary(
                '/xgtproto.AdminService/MaxUserMemorySize',
                request_serializer=AdminService__pb2.MaxUserMemorySizeRequest.SerializeToString,
                response_deserializer=AdminService__pb2.MaxUserMemorySizeResponse.FromString,
                _registered_method=True)
        self.Shutdown = channel.unary_unary(
                '/xgtproto.AdminService/Shutdown',
                request_serializer=AdminService__pb2.ShutdownRequest.SerializeToString,
                response_deserializer=AdminService__pb2.ShutdownResponse.FromString,
                _registered_method=True)
        self.Version = channel.unary_unary(
                '/xgtproto.AdminService/Version',
                request_serializer=AdminService__pb2.VersionRequest.SerializeToString,
                response_deserializer=AdminService__pb2.VersionResponse.FromString,
                _registered_method=True)
        self.Authenticate = channel.unary_unary(
                '/xgtproto.AdminService/Authenticate',
                request_serializer=AdminService__pb2.AuthenticateRequest.SerializeToString,
                response_deserializer=AdminService__pb2.AuthenticateResponse.FromString,
                _registered_method=True)
        self.PKIAuthenticate = channel.unary_unary(
                '/xgtproto.AdminService/PKIAuthenticate',
                request_serializer=AdminService__pb2.AuthenticateRequest.SerializeToString,
                response_deserializer=AdminService__pb2.AuthenticateResponse.FromString,
                _registered_method=True)
        self.ProxyPKIAuthenticate = channel.unary_unary(
                '/xgtproto.AdminService/ProxyPKIAuthenticate',
                request_serializer=AdminService__pb2.AuthenticateRequest.SerializeToString,
                response_deserializer=AdminService__pb2.AuthenticateResponse.FromString,
                _registered_method=True)
        self.KerberosAuthenticate = channel.unary_unary(
                '/xgtproto.AdminService/KerberosAuthenticate',
                request_serializer=AdminService__pb2.KerberosAuthenticateRequest.SerializeToString,
                response_deserializer=AdminService__pb2.KerberosAuthenticateResponse.FromString,
                _registered_method=True)
        self.CloseSession = channel.unary_unary(
                '/xgtproto.AdminService/CloseSession',
                request_serializer=AdminService__pb2.CloseSessionRequest.SerializeToString,
                response_deserializer=AdminService__pb2.CloseSessionResponse.FromString,
                _registered_method=True)
        self.KeepAliveSession = channel.unary_unary(
                '/xgtproto.AdminService/KeepAliveSession',
                request_serializer=AdminService__pb2.KeepAliveSessionRequest.SerializeToString,
                response_deserializer=AdminService__pb2.KeepAliveSessionResponse.FromString,
                _registered_method=True)
        self.GetConfig = channel.unary_unary(
                '/xgtproto.AdminService/GetConfig',
                request_serializer=AdminService__pb2.GetConfigRequest.SerializeToString,
                response_deserializer=AdminService__pb2.GetConfigResponse.FromString,
                _registered_method=True)
        self.SetConfig = channel.unary_unary(
                '/xgtproto.AdminService/SetConfig',
                request_serializer=AdminService__pb2.SetConfigRequest.SerializeToString,
                response_deserializer=AdminService__pb2.SetConfigResponse.FromString,
                _registered_method=True)
        self.SetDefaultNamespace = channel.unary_unary(
                '/xgtproto.AdminService/SetDefaultNamespace',
                request_serializer=AdminService__pb2.SetDefaultNamespaceRequest.SerializeToString,
                response_deserializer=AdminService__pb2.SetDefaultNamespaceResponse.FromString,
                _registered_method=True)
        self.GetDefaultNamespace = channel.unary_unary(
                '/xgtproto.AdminService/GetDefaultNamespace',
                request_serializer=AdminService__pb2.GetDefaultNamespaceRequest.SerializeToString,
                response_deserializer=AdminService__pb2.GetDefaultNamespaceResponse.FromString,
                _registered_method=True)
        self.IsAdmin = channel.unary_unary(
                '/xgtproto.AdminService/IsAdmin',
                request_serializer=AdminService__pb2.IsAdminRequest.SerializeToString,
                response_deserializer=AdminService__pb2.IsAdminResponse.FromString,
                _registered_method=True)


class AdminServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def ChangeErrorCount(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def FreeUserMemorySize(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def MaxUserMemorySize(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Shutdown(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Version(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Authenticate(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def PKIAuthenticate(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def ProxyPKIAuthenticate(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def KerberosAuthenticate(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CloseSession(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def KeepAliveSession(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetConfig(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SetConfig(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SetDefaultNamespace(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def GetDefaultNamespace(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def IsAdmin(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_AdminServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'ChangeErrorCount': grpc.unary_unary_rpc_method_handler(
                    servicer.ChangeErrorCount,
                    request_deserializer=AdminService__pb2.ChangeErrorCountRequest.FromString,
                    response_serializer=AdminService__pb2.ChangeErrorCountResponse.SerializeToString,
            ),
            'FreeUserMemorySize': grpc.unary_unary_rpc_method_handler(
                    servicer.FreeUserMemorySize,
                    request_deserializer=AdminService__pb2.FreeUserMemorySizeRequest.FromString,
                    response_serializer=AdminService__pb2.FreeUserMemorySizeResponse.SerializeToString,
            ),
            'MaxUserMemorySize': grpc.unary_unary_rpc_method_handler(
                    servicer.MaxUserMemorySize,
                    request_deserializer=AdminService__pb2.MaxUserMemorySizeRequest.FromString,
                    response_serializer=AdminService__pb2.MaxUserMemorySizeResponse.SerializeToString,
            ),
            'Shutdown': grpc.unary_unary_rpc_method_handler(
                    servicer.Shutdown,
                    request_deserializer=AdminService__pb2.ShutdownRequest.FromString,
                    response_serializer=AdminService__pb2.ShutdownResponse.SerializeToString,
            ),
            'Version': grpc.unary_unary_rpc_method_handler(
                    servicer.Version,
                    request_deserializer=AdminService__pb2.VersionRequest.FromString,
                    response_serializer=AdminService__pb2.VersionResponse.SerializeToString,
            ),
            'Authenticate': grpc.unary_unary_rpc_method_handler(
                    servicer.Authenticate,
                    request_deserializer=AdminService__pb2.AuthenticateRequest.FromString,
                    response_serializer=AdminService__pb2.AuthenticateResponse.SerializeToString,
            ),
            'PKIAuthenticate': grpc.unary_unary_rpc_method_handler(
                    servicer.PKIAuthenticate,
                    request_deserializer=AdminService__pb2.AuthenticateRequest.FromString,
                    response_serializer=AdminService__pb2.AuthenticateResponse.SerializeToString,
            ),
            'ProxyPKIAuthenticate': grpc.unary_unary_rpc_method_handler(
                    servicer.ProxyPKIAuthenticate,
                    request_deserializer=AdminService__pb2.AuthenticateRequest.FromString,
                    response_serializer=AdminService__pb2.AuthenticateResponse.SerializeToString,
            ),
            'KerberosAuthenticate': grpc.unary_unary_rpc_method_handler(
                    servicer.KerberosAuthenticate,
                    request_deserializer=AdminService__pb2.KerberosAuthenticateRequest.FromString,
                    response_serializer=AdminService__pb2.KerberosAuthenticateResponse.SerializeToString,
            ),
            'CloseSession': grpc.unary_unary_rpc_method_handler(
                    servicer.CloseSession,
                    request_deserializer=AdminService__pb2.CloseSessionRequest.FromString,
                    response_serializer=AdminService__pb2.CloseSessionResponse.SerializeToString,
            ),
            'KeepAliveSession': grpc.unary_unary_rpc_method_handler(
                    servicer.KeepAliveSession,
                    request_deserializer=AdminService__pb2.KeepAliveSessionRequest.FromString,
                    response_serializer=AdminService__pb2.KeepAliveSessionResponse.SerializeToString,
            ),
            'GetConfig': grpc.unary_unary_rpc_method_handler(
                    servicer.GetConfig,
                    request_deserializer=AdminService__pb2.GetConfigRequest.FromString,
                    response_serializer=AdminService__pb2.GetConfigResponse.SerializeToString,
            ),
            'SetConfig': grpc.unary_unary_rpc_method_handler(
                    servicer.SetConfig,
                    request_deserializer=AdminService__pb2.SetConfigRequest.FromString,
                    response_serializer=AdminService__pb2.SetConfigResponse.SerializeToString,
            ),
            'SetDefaultNamespace': grpc.unary_unary_rpc_method_handler(
                    servicer.SetDefaultNamespace,
                    request_deserializer=AdminService__pb2.SetDefaultNamespaceRequest.FromString,
                    response_serializer=AdminService__pb2.SetDefaultNamespaceResponse.SerializeToString,
            ),
            'GetDefaultNamespace': grpc.unary_unary_rpc_method_handler(
                    servicer.GetDefaultNamespace,
                    request_deserializer=AdminService__pb2.GetDefaultNamespaceRequest.FromString,
                    response_serializer=AdminService__pb2.GetDefaultNamespaceResponse.SerializeToString,
            ),
            'IsAdmin': grpc.unary_unary_rpc_method_handler(
                    servicer.IsAdmin,
                    request_deserializer=AdminService__pb2.IsAdminRequest.FromString,
                    response_serializer=AdminService__pb2.IsAdminResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'xgtproto.AdminService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('xgtproto.AdminService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class AdminService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def ChangeErrorCount(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/xgtproto.AdminService/ChangeErrorCount',
            AdminService__pb2.ChangeErrorCountRequest.SerializeToString,
            AdminService__pb2.ChangeErrorCountResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def FreeUserMemorySize(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/xgtproto.AdminService/FreeUserMemorySize',
            AdminService__pb2.FreeUserMemorySizeRequest.SerializeToString,
            AdminService__pb2.FreeUserMemorySizeResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def MaxUserMemorySize(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/xgtproto.AdminService/MaxUserMemorySize',
            AdminService__pb2.MaxUserMemorySizeRequest.SerializeToString,
            AdminService__pb2.MaxUserMemorySizeResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Shutdown(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/xgtproto.AdminService/Shutdown',
            AdminService__pb2.ShutdownRequest.SerializeToString,
            AdminService__pb2.ShutdownResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Version(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/xgtproto.AdminService/Version',
            AdminService__pb2.VersionRequest.SerializeToString,
            AdminService__pb2.VersionResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def Authenticate(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/xgtproto.AdminService/Authenticate',
            AdminService__pb2.AuthenticateRequest.SerializeToString,
            AdminService__pb2.AuthenticateResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def PKIAuthenticate(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/xgtproto.AdminService/PKIAuthenticate',
            AdminService__pb2.AuthenticateRequest.SerializeToString,
            AdminService__pb2.AuthenticateResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def ProxyPKIAuthenticate(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/xgtproto.AdminService/ProxyPKIAuthenticate',
            AdminService__pb2.AuthenticateRequest.SerializeToString,
            AdminService__pb2.AuthenticateResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def KerberosAuthenticate(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/xgtproto.AdminService/KerberosAuthenticate',
            AdminService__pb2.KerberosAuthenticateRequest.SerializeToString,
            AdminService__pb2.KerberosAuthenticateResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def CloseSession(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/xgtproto.AdminService/CloseSession',
            AdminService__pb2.CloseSessionRequest.SerializeToString,
            AdminService__pb2.CloseSessionResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def KeepAliveSession(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/xgtproto.AdminService/KeepAliveSession',
            AdminService__pb2.KeepAliveSessionRequest.SerializeToString,
            AdminService__pb2.KeepAliveSessionResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetConfig(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/xgtproto.AdminService/GetConfig',
            AdminService__pb2.GetConfigRequest.SerializeToString,
            AdminService__pb2.GetConfigResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def SetConfig(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/xgtproto.AdminService/SetConfig',
            AdminService__pb2.SetConfigRequest.SerializeToString,
            AdminService__pb2.SetConfigResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def SetDefaultNamespace(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/xgtproto.AdminService/SetDefaultNamespace',
            AdminService__pb2.SetDefaultNamespaceRequest.SerializeToString,
            AdminService__pb2.SetDefaultNamespaceResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def GetDefaultNamespace(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/xgtproto.AdminService/GetDefaultNamespace',
            AdminService__pb2.GetDefaultNamespaceRequest.SerializeToString,
            AdminService__pb2.GetDefaultNamespaceResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def IsAdmin(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/xgtproto.AdminService/IsAdmin',
            AdminService__pb2.IsAdminRequest.SerializeToString,
            AdminService__pb2.IsAdminResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from . import DataService_pb2 as DataService__pb2


class DataServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.UploadData = channel.stream_unary(
                '/xgtproto.DataService/UploadData',
                request_serializer=DataService__pb2.UploadDataRequest.SerializeToString,
                response_deserializer=DataService__pb2.UploadDataResponse.FromString,
                _registered_method=True)
        self.DownloadData = channel.unary_stream(
                '/xgtproto.DataService/DownloadData',
                request_serializer=DataService__pb2.DownloadDataRequest.SerializeToString,
                response_deserializer=DataService__pb2.DownloadDataResponse.FromString,
                _registered_method=True)
        self.IngestUri = channel.stream_unary(
                '/xgtproto.DataService/IngestUri',
                request_serializer=DataService__pb2.IngestUriRequest.SerializeToString,
                response_deserializer=DataService__pb2.IngestUriResponse.FromString,
                _registered_method=True)
        self.EgestUri = channel.unary_unary(
                '/xgtproto.DataService/EgestUri',
                request_serializer=DataService__pb2.EgestUriRequest.SerializeToString,
                response_deserializer=DataService__pb2.EgestUriResponse.FromString,
                _registered_method=True)
        self.InferSchema = channel.unary_unary(
                '/xgtproto.DataService/InferSchema',
                request_serializer=DataService__pb2.InferSchemaRequest.SerializeToString,
                response_deserializer=DataService__pb2.InferSchemaResponse.FromString,
                _registered_method=True)
        self.TransferFrameArrowClient = channel.unary_unary(
                '/xgtproto.DataService/TransferFrameArrowClient',
                request_serializer=DataService__pb2.TransferFrameArrowClientRequest.SerializeToString,
                response_deserializer=DataService__pb2.TransferFrameArrowClientResponse.FromString,
                _registered_method=True)
        self.UploadCSVMultiData = channel.stream_unary(
                '/xgtproto.DataService/UploadCSVMultiData',
                request_serializer=DataService__pb2.UploadCSVMultiDataRequest.SerializeToString,
                response_deserializer=DataService__pb2.UploadCSVMultiDataResponse.FromString,
                _registered_method=True)
        self.UploadArrowMultiData = channel.stream_unary(
                '/xgtproto.DataService/UploadArrowMultiData',
                request_serializer=DataService__pb2.UploadArrowMultiDataRequest.SerializeToString,
                response_deserializer=DataService__pb2.UploadArrowMultiDataResponse.FromString,
                _registered_method=True)


class DataServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def UploadData(self, request_iterator, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def DownloadData(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def IngestUri(self, request_iterator, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def EgestUri(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def InferSchema(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def TransferFrameArrowClient(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def UploadCSVMultiData(self, request_iterator, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def UploadArrowMultiData(self, request_iterator, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_DataServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'UploadData': grpc.stream_unary_rpc_method_handler(
                    servicer.UploadData,
                    request_deserializer=DataService__pb2.UploadDataRequest.FromString,
                    response_serializer=DataService__pb2.UploadDataResponse.SerializeToString,
            ),
            'DownloadData': grpc.unary_stream_rpc_method_handler(
                    servicer.DownloadData,
                    request_deserializer=DataService__pb2.DownloadDataRequest.FromString,
                    response_serializer=DataService__pb2.DownloadDataResponse.SerializeToString,
            ),
            'IngestUri': grpc.stream_unary_rpc_method_handler(
                    servicer.IngestUri,
                    request_deserializer=DataService__pb2.IngestUriRequest.FromString,
                    response_serializer=DataService__pb2.IngestUriResponse.SerializeToString,
            ),
            'EgestUri': grpc.unary_unary_rpc_method_handler(
                    servicer.EgestUri,
                    request_deserializer=DataService__pb2.EgestUriRequest.FromString,
                    response_serializer=DataService__pb2.EgestUriResponse.SerializeToString,
            ),
            'InferSchema': grpc.unary_unary_rpc_method_handler(
                    servicer.InferSchema,
                    request_deserializer=DataService__pb2.InferSchemaRequest.FromString,
                    response_serializer=DataService__pb2.InferSchemaResponse.SerializeToString,
            ),
            'TransferFrameArrowClient': grpc.unary_unary_rpc_method_handler(
                    servicer.TransferFrameArrowClient,
                    request_deserializer=DataService__pb2.TransferFrameArrowClientRequest.FromString,
                    response_serializer=DataService__pb2.TransferFrameArrowClientResponse.SerializeToString,
            ),
            'UploadCSVMultiData': grpc.stream_unary_rpc_method_handler(
                    servicer.UploadCSVMultiData,
                    request_deserializer=DataService__pb2.UploadCSVMultiDataRequest.FromString,
                    response_serializer=DataService__pb2.UploadCSVMultiDataResponse.SerializeToString,
            ),
            'UploadArrowMultiData': grpc.stream_unary_rpc_method_handler(
                    servicer.UploadArrowMultiData,
                    request_deserializer=DataService__pb2.UploadArrowMultiDataRequest.FromString,
                    response_serializer=DataService__pb2.UploadArrowMultiDataResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'xgtproto.DataService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))
    server.add_registered_method_handlers('xgtproto.DataService', rpc_method_handlers)


 # This class is part of an EXPERIMENTAL API.
class DataService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def UploadData(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_unary(
            request_iterator,
            target,
            '/xgtproto.DataService/UploadData',
            DataService__pb2.UploadDataRequest.SerializeToString,
            DataService__pb2.UploadDataResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def DownloadData(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(
            request,
            target,
            '/xgtproto.DataService/DownloadData',
            DataService__pb2.DownloadDataRequest.SerializeToString,
            DataService__pb2.DownloadDataResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def IngestUri(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_unary(
            request_iterator,
            target,
            '/xgtproto.DataService/IngestUri',
            DataService__pb2.IngestUriRequest.SerializeToString,
            DataService__pb2.IngestUriResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def EgestUri(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/xgtproto.DataService/EgestUri',
            DataService__pb2.EgestUriRequest.SerializeToString,
            DataService__pb2.EgestUriResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def InferSchema(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/xgtproto.DataService/InferSchema',
            DataService__pb2.InferSchemaRequest.SerializeToString,
            DataService__pb2.InferSchemaResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def TransferFrameArrowClient(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(
            request,
            target,
            '/xgtproto.DataService/TransferFrameArrowClient',
            DataService__pb2.TransferFrameArrowClientRequest.SerializeToString,
            DataService__pb2.TransferFrameArrowClientResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def UploadCSVMultiData(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_unary(
            request_iterator,
            target,
            '/xgtproto.DataService/UploadCSVMultiData',
            DataService__pb2.UploadCSVMultiDataRequest.SerializeToString,
            DataService__pb2.UploadCSVMultiDataResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)

    @staticmethod
    def UploadArrowMultiData(request_iterator,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.stream_unary(
            request_iterator,
            target,
            '/xgtproto.DataService/UploadArrowMultiData',
            DataService__pb2.UploadArrowMultiDataRequest.SerializeToString,
            DataService__pb2.UploadArrowMultiDataResponse.FromString,
            options,
            channel_credentials,
            insecure,
            call_credentials,
            compression,
            wait_for_ready,
            timeout,
            metadata,
            _registered_method=True)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# NO CHECKED-IN PROTOBUF GENCODE
# source: JobService.proto
# Protobuf Python Version: 5.28.1
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import runtime_version as _runtime_version
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder
_runtime_version.ValidateProtobufRuntimeVersion(
    _runtime_version.Domain.PUBLIC,
    5,
    28,
    1,
    '',
    'JobService.proto'
)
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from . import AdminService_pb2 as AdminService__pb2
from . import ErrorMessages_pb2 as ErrorMessages__pb2
from . import SchemaMessages_pb2 as SchemaMessages__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(b'\n\x10JobService.proto\x12\x08xgtproto\x1a\x12\x41\x64minService.proto\x1a\x13\x45rrorMessages.proto\x1a\x14SchemaMessages.proto\"\xce\x02\n\x19ScheduleJobsCypherRequest\x12\x14\n\x0c\x63ypher_query\x18\x01 \x03(\t\x12?\n\x06kwargs\x18\x05 \x03(\x0b\x32/.xgtproto.ScheduleJobsCypherRequest.KwargsEntry\x12G\n\nparameters\x18\x06 \x03(\x0b\x32\x33.xgtproto.ScheduleJobsCypherRequest.ParametersEntry\x1a\x44\n\x0bKwargsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12$\n\x05value\x18\x02 \x01(\x0b\x32\x15.xgtproto.ConfigValue:\x02\x38\x01\x1aK\n\x0fParametersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\'\n\x05value\x18\x02 \x01(\x0b\x32\x18.xgtproto.ParameterValue:\x02\x38\x01\"e\n\x1aScheduleJobsCypherResponse\x12\x1e\n\x05\x65rror\x18\x01 \x03(\x0b\x32\x0f.xgtproto.Error\x12\'\n\njob_status\x18\x02 \x03(\x0b\x32\x13.xgtproto.JobStatus\"#\n\x11\x43\x61ncelJobsRequest\x12\x0e\n\x06job_id\x18\x01 \x03(\x04\"]\n\x12\x43\x61ncelJobsResponse\x12\x1e\n\x05\x65rror\x18\x01 \x03(\x0b\x32\x0f.xgtproto.Error\x12\'\n\njob_status\x18\x02 \x03(\x0b\x32\x13.xgtproto.JobStatus\" \n\x0eGetJobsRequest\x12\x0e\n\x06job_id\x18\x01 \x03(\x04\"Z\n\x0fGetJobsResponse\x12\x1e\n\x05\x65rror\x18\x01 \x03(\x0b\x32\x0f.xgtproto.Error\x12\'\n\njob_status\x18\x02 \x01(\x0b\x32\x13.xgtproto.JobStatus\"2\n\x0fWaitJobsRequest\x12\x0e\n\x06job_id\x18\x01 \x03(\x04\x12\x0f\n\x07timeout\x18\x02 \x01(\x03\"[\n\x10WaitJobsResponse\x12\x1e\n\x05\x65rror\x18\x01 \x03(\x0b\x32\x0f.xgtproto.Error\x12\'\n\njob_status\x18\x02 \x03(\x0b\x32\x13.xgtproto.JobStatus\"\xb5\x02\n\x13\x43ompileQueryRequest\x12\r\n\x05query\x18\x01 \x01(\t\x12\x39\n\x06kwargs\x18\x02 \x03(\x0b\x32).xgtproto.CompileQueryRequest.KwargsEntry\x12\x41\n\nparameters\x18\x03 \x03(\x0b\x32-.xgtproto.CompileQueryRequest.ParametersEntry\x1a\x44\n\x0bKwargsEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12$\n\x05value\x18\x02 \x01(\x0b\x32\x15.xgtproto.ConfigValue:\x02\x38\x01\x1aK\n\x0fParametersEntry\x12\x0b\n\x03key\x18\x01 \x01(\t\x12\'\n\x05value\x18\x02 \x01(\x0b\x32\x18.xgtproto.ParameterValue:\x02\x38\x01\"6\n\x14\x43ompileQueryResponse\x12\x1e\n\x05\x65rror\x18\x01 \x03(\x0b\x32\x0f.xgtproto.Error2\x94\x03\n\nJobService\x12\x61\n\x12ScheduleJobsCypher\x12#.xgtproto.ScheduleJobsCypherRequest\x1a$.xgtproto.ScheduleJobsCypherResponse\"\x00\x12I\n\nCancelJobs\x12\x1b.xgtproto.CancelJobsRequest\x1a\x1c.xgtproto.CancelJobsResponse\"\x00\x12\x42\n\x07GetJobs\x12\x18.xgtproto.GetJobsRequest\x1a\x19.xgtproto.GetJobsResponse\"\x00\x30\x01\x12\x43\n\x08WaitJobs\x12\x19.xgtproto.WaitJobsRequest\x1a\x1a.xgtproto.WaitJobsResponse\"\x00\x12O\n\x0c\x43ompileQuery\x12\x1d.xgtproto.CompileQueryRequest\x1a\x1e.xgtproto.CompileQueryResponse\"\x00\x62\x06proto3')

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(DESCRIPTOR, 'JobService_pb2', _globals)
if not _descriptor._USE_C_DESCRIPTORS:
  DESCRIPTOR._loaded_options = None
  _globals['_SCHEDULEJOBSCYPHERREQUEST_KWARGSENTRY']._loaded_options = None
  _globals['_SCHEDULEJOBSCYPHERREQUEST_KWARGSENTRY']._serialized_options = b'8\001'
  _globals['_SCHEDULEJOBSCYPHERREQUEST_PARAMETERSENTRY']._loaded_options = None
  _globals['_SCHEDULEJOBSCYPHERREQUEST_PARAMETERSENTRY']._serialized_options = b'8\001'
  _globals['_COMPILEQUERYREQUEST_KWARGSENTRY']._loaded_options = None
  _globals['_COMPILEQUERYREQUEST_KWARGSENTRY']._serialized_options = b'8\001'
  _globals['_COMPILEQUERYREQUEST_PARAMETERSENTRY']._loaded_options = None
  _globals['_COMPILEQUERYREQUEST_PARAMETERSENTRY']._serialized_options = b'8\001'
  _globals['_SCHEDULEJOBSCYPHERREQUEST']._serialized_start=94
  _globals['_SCHEDULEJOBSCYPHERREQUEST']._serialized_end=428
  _globals['_SCHEDULEJOBSCYPHERREQUEST_KWARGSENTRY']._serialized_start=283
  _globals['_SCHEDULEJOBSCYPHERREQUEST_KWARGSENTRY']._serialized_end=351
  _globals['_SCHEDULEJOBSCYPHERREQUEST_PARAMETERSENTRY']._serialized_start=353
  _globals['_SCHEDULEJOBSCYPHERREQUEST_PARAMETERSENTRY']._serialized_end=428
  _globals['_SCHEDULEJOBSCYPHERRESPONSE']._serialized_start=430
  _globals['_SCHEDULEJOBSCYPHERRESPONSE']._serialized_end=531
  _globals['_CANCELJOBSREQUEST']._serialized_start=533
  _globals['_CANCELJOBSREQUEST']._serialized_end=568
  _globals['_CANCELJOBSRESPONSE']._serialized_start=570
  _globals['_CANCELJOBSRESPONSE']._serialized_end=663
  _globals['_GETJOBSREQUEST']._serialized_start=665
  _globals['_GETJOBSREQUEST']._serialized_end=697
  _globals['_GETJOBSRESPONSE']._serialized_start=699
  _globals['_GETJOBSRESPONSE']._serialized_end=789
  _globals['_WAITJOBSREQUEST']._serialized_start=791
  _globals['_WAITJOBSREQUEST']._serialized_end=841
  _globals['_WAITJOBSRESPONSE']._serialized_start=843
  _globals['_WAITJOBSRESPONSE']._serialized_end=934
  _globals['_COMPILEQUERYREQUEST']._serialized_start=937
  _globals['_COMPILEQUERYREQUEST']._serialized_end=1246
  _globals['_COMPILEQUERYREQUEST_KWARGSENTRY']._serialized_start=283
  _globals['_COMPILEQUERYREQUEST_KWARGSENTRY']._serialized_end=351
  _globals['_COMPILEQUERYREQUEST_PARAMETERSENTRY']._serialized_start=353
  _globals['_COMPILEQUERYREQUEST_PARAMETERSENTRY']._serialized_end=428
  _globals['_COMPILEQUERYRESPONSE']._serialized_start=1248
  _globals['_COMPILEQUERYRESPONSE']._serialized_end=1302
  _globals['_JOBSERVICE']._serialized_start=1305
  _globals['_JOBSERVICE']._serialized_end=1709
# @@protoc_insertion_point(module_scope)



# -*- coding: utf-8 -*- ----------------------------------------------------===#
#
#  Copyright 2016-2025 Trovares Inc. dba Rocketgraph.  All rights reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.
#
#===------------------------------------------------------------------------===#

from __future__ import annotations

import json
import logging
import pyarrow
import sys

from collections.abc import Iterable, Mapping, Sequence

# In Python 3.9 and above, the abstract base classes (ABCs) from collections.abc
# can be used directly as type hints. This includes Iterable, Mapping, and
# Sequence, among others. For compatibility with versions below 3.9, we define
# aliases using the typing module, which has equivalent types with the same
# names. This ensures our code can use consistent type hints across different
# Python versions.
if sys.version_info >= (3, 9):
  # Directly use the ABCs from collections.abc in type hints since it is
  # supported. List and Dict type hints are directly available as built-in types
  # in Python 3.9+.
  Iter, Map, Seq, List, Dict = Iterable, Mapping, Sequence, list, dict
else:
   # For Python versions below 3.9, use the typing module for type hints.
  from typing import (Iterable as Iter, Mapping as Map, Sequence as Seq, List,
                                  Dict)

from datetime import datetime
from typing import Optional, Union, TYPE_CHECKING
from . import ErrorMessages_pb2 as err_proto
from . import JobService_pb2 as job_proto
from . import SchemaMessages_pb2 as sch_proto
from .common import (XgtError, XgtInternalError, XgtNotImplemented,
                     XgtTypeError, XgtValueError, XgtErrorTypes,
                     _create_flight_ticket, _verify_offset_length,
                     _validated_columns, _get_data_python_from_table,
                     _get_data_pandas_from_table,
                     _get_data_arrow, _code_error_map, _deprecated)

if TYPE_CHECKING:
  try:
    import pandas
  except ImportError:
    pass

log = logging.getLogger(__name__)

class _QueryPlan(object):
  class _QueryPlanSingleFrame(object):
    def __init__(self, query_element_message):
      self._label = query_element_message.source_label
      self._frame = query_element_message.source_frame

    @property
    def name(self):
      return f"{self._label}:{self._frame}"

    def __str__(self):
      return f"({self._label}:{self._frame})"

  class _QueryPlanEdge(object):
    def __init__(self, query_edge_message):
      if (query_edge_message.edge_direction ==
            sch_proto.QueryEdgeDirection.Value('FORWARD')):
        self.forward_edge = True
      else:
        self.forward_edge = False
      self._source_label = query_edge_message.source_label
      self._source_frame = query_edge_message.source_frame
      self._edge_label = query_edge_message.edge_label
      self._edge_frame = query_edge_message.edge_frame
      self._target_label = query_edge_message.target_label
      self._target_frame = query_edge_message.target_frame

    @property
    def name(self):
      return f"{self._edge_label}:{self._edge_frame}"

    @property
    def edge_name(self):
      return f"{self._edge_label}:{self._edge_frame}"

    @property
    def source_name(self):
      return f"{self._source_label}:{self._source_frame}"

    @property
    def target_name(self):
      return f"{self._target_label}:{self._target_frame}"

    # The string representation of a _QueryPlanEdge object is a tuple
    # of the source vertex, edge, and target vertex. The direction
    # of the edge is always forward.
    def __str__(self):
      if self.forward_edge:
        vertex1 = self.source_name
        vertex2 = self.target_name
      else:
        vertex2 = self.source_name
        vertex1 = self.target_name
      return f"({vertex1}, {self.edge_name}, {vertex2})"

  def __init__(self, query_plan_response):
    # The _plan member is a list of lists. Each list represents a query
    # that starts with MATCH in cypher and contains elements representing
    # edges or (if only a single frame was matched), a vertex or table.
    self._plan = []
    for query in query_plan_response.plan_as_edge_list:
      query_representation = []
      plan_score = None
      for element in query.edges:
        if (element.edge_direction ==
              sch_proto.QueryEdgeDirection.Value('NOT_EDGE')):
          query_representation.append(self._QueryPlanSingleFrame(element))
        else:
          query_representation.append(self._QueryPlanEdge(element))
      if query.HasField("plan_score"):
        plan_score = query.plan_score
      self._plan.append((query_representation, plan_score))

  def __str__(self):
    as_string = ""
    for (idx, query) in enumerate(self._plan):
      if idx > 0:
        as_string += "\n"
      if len(query[0]) > 0:
        as_string += "QUERY:"
        for element in query[0]:
          as_string += f"\n{element}"
    return as_string

  def to_cypher(self):
    cypher = ""
    for query, score in self._plan:
      # The corresponding query section had no MATCH portion. It won't be
      # included in the cypher representation currently used as only
      # the structure of the MATCH portions is shown.
      if len(query) == 0:
        continue
      if len(cypher) > 0:
        cypher += "\n"
      cypher += "MATCH\n"
      last_target_name = ""

      # Write the first element of the query. This must be a vertex or table
      # frame.
      if isinstance(query[0], self._QueryPlanEdge):
        cypher += f"({query[0].source_name})"
      else:
        cypher += f"({query[0].name})"

      for (element_idx, element) in enumerate(query):
        if isinstance(element, self._QueryPlanEdge):
          if last_target_name != element.source_name and element_idx > 0:
            # This edge starts a new path in the query. We must write the
            # source.
            cypher += f",\n({element.source_name})"
          if not element.forward_edge:
            cypher += "<"
          cypher += f"-[{element.edge_name}]-"
          if element.forward_edge:
            cypher += ">"
          cypher += f"({element.target_name})"
          last_target_name = element.target_name
        elif element_idx > 0:
          # This is not an edge element. It is a vertex or table frame.
          # If it was the first in this query then it has already been
          # written. Otherwise, it starts a new path in the query and
          # must be written.
          cypher += f",\n({element.name})"
    return cypher

  def get_scores(self):
    return [score for query, score in self._plan]

class Job(object):
  """
  Represents a user-scheduled Job.

  An instance of this object is created by job-scheduling functions like
  `xgt.Connection.run_job` and `xgt.Connection.schedule_job`.

  A `Job` is used as a proxy for a job in the server and allows the user
  to monitor its execution, possibly cancel it, and learn about its status
  during and after execution.

  The conn parameter represents a previously established connection to the xGT
  server.

  The job_response parameter is a single element of the array returned by the
  output of a job creation gRPC call.  Each individual element in the array will
  be constructed as a separate `Job` object.
  """
  def __init__(self, conn, job_response = None, python_errors = None):
    """
    Constructor for Job. Called when Job is created.
    """
    self._conn = conn
    if job_response is not None:
      self._id = job_response.job_id
      self._user = job_response.user
      self._data = self._parse_job_data(job_response)
    else:
      self._id = 0
      self._user = ""
      self._data = { 'status' : 'completed', 'num_rows' : 0 }

    if python_errors is not None:
      if 'ingest_errors' in self._data.keys():
        self._data['ingest_errors'] += python_errors
        self._data['total_ingest_errors'] += len(python_errors)
      else:
        if len(python_errors) > 0:
          self._data['ingest_errors'] = python_errors
        self._data['total_ingest_errors'] = len(python_errors)

  def _parse_job_data(self, job_response):
    job_data = {
      'jobid': job_response.job_id,
      'user': job_response.user,
      'status': sch_proto.JobStatusEnum.Name(job_response.status).lower(),
      'start_time': job_response.start_time.ToDatetime().isoformat(),
      'end_time': job_response.end_time.ToDatetime().isoformat(),
      'error_type': None,
      'visited_edges': job_response.visited_edges,
      'timing': job_response.timing,
      'description': job_response.description,
      'num_rows' : job_response.num_rows,
      'total_ingest_errors' : job_response.total_ingest_errors,
      'default_namespace': job_response.default_namespace,
      'results_frame': job_response.results_frame,
    }
    # Optional fields.
    if job_response.HasField('schema'):
      schema = self._conn._translate_schema_from_server(job_response.schema)
      job_data['schema'] = schema
    if job_response.ingest_error and len(job_response.ingest_error):
      job_data['ingest_errors'] = [data for data in job_response.ingest_error]
    if job_response.error and len(job_response.error) > 0:
        error_code_name = err_proto.ErrorCodeEnum.Name(
            job_response.error[0].code)
        job_data['error_type'] = _code_error_map[error_code_name]
        job_data['error'] = ', '.join([e.message for e in job_response.error])
        job_data['trace'] = ', '.join([e.detail for e in job_response.error])
    if job_response.HasField('query_plan'):
      job_data['query_plan'] =  _QueryPlan(job_response.query_plan)

    if job_response.end_time.ToSeconds() == 0:
      job_data['end_time'] = ''

    return job_data

  def _get_job_data(self):
    request = job_proto.GetJobsRequest()
    request.job_id.extend([self._id])
    responses = self._conn._call(request, self._conn._job_svc.GetJobs)
    job_data = None
    resp_cnt = 0
    for response in responses: # Expect only one response.
      # Retrieve the job status.
      job_data = self._parse_job_data(response.job_status)
      resp_cnt += 1

    if resp_cnt > 1:
      raise XgtInternalError("Expected a single job in response")

    # If the status is unknown, only update the status, not other fields as
    # they will be invalid.
    returned_status = job_data['status']
    if returned_status == 'unknown_job_status':
      self._data['status'] = returned_status
      return job_data

    # Cache the response from the server.
    self._data = job_data

    if log.getEffectiveLevel() >= logging.DEBUG:
      job_id = job_data['jobid']
      job_status = job_data['status']
      user = job_data['user']
      if 'error' in job_data:
        error = job_data['error']
      else:
        error = ''
      if 'trace' in job_data:
        trace = job_data['trace']
      else:
        trace = ''
      msg = f'Job: {job_id} User: {user} Status: {job_status}'
      if error != '':
        msg += "\nError: \n" + error
      if trace != '':
        msg += "\nTrace: \n" + trace
      log.debug(msg)

    return job_data

  def _is_status_final(self):
    if 'status' in self._data:
      curr_status = self._data['status']

      if ((curr_status == 'completed') or (curr_status == 'canceled') or
          (curr_status == 'failed') or (curr_status == 'rollback') or
          (curr_status == 'unknown_job_status')):
        return True

    return False

  # Returns the job results in arrow table form taking into account the
  # requested offset and length.  If the results are not yet available, None is
  # returned.
  def _load_arrow_table(self, offset, length, rows=None, columns=None):
    if not self._is_status_final():
      self._get_job_data()

    if self._is_status_final():
      curr_status = self._data['status']
      if curr_status != 'completed':
        return None

      # Skip getting the Arrow table if there are no results.  The schema is
      # not set for jobs without immediate results.
      if 'schema' not in self._data:
        return None

      ticket = _create_flight_ticket(self, 'xgt__Job_History',
                                     offset, length,
                                     rows = rows, columns = columns,
                                     include_row_labels = False,
                                     job_id = self._id)

      return _get_data_arrow(self._conn, ticket)

    return None

  @property
  def id(self) -> int:
    """
    int: Identifier of the job.

    A 64-bit integer value that uniquely identifies a job. It is
    automatically incremented for each scheduled job over the lifetime of
    the xGT server process.
    """
    return self._id

  @property
  def user(self) -> str:
    """
    str: User who ran the job.
    """
    return self._user

  @property
  def status(self) -> str:
    """
    str: Status of the job.

    ==================  ===============================================
          Job status
    -------------------------------------------------------------------
             Status                       Description
    ==================  ===============================================
             scheduled  The state after the job has been created, but
                        before it has started running.
               running  The job is being executed.
             completed  The job finished successfully.
              canceled  The job was canceled.
                failed  The job failed. When the job fails the `error`
                        and `trace` properties are populated.
              rollback  The job had a transactional conflict with
                        another job and was rolled back.
    unknown_job_status  The job was not found in the job history.
    ==================  ===============================================
    """
    if not self._is_status_final():
      self._get_job_data()

    if 'status' in self._data:
      return self._data['status']
    else:
      return ''

  @property
  def start_time(self) -> str:
    """
    str: Date and time when the job was scheduled.

    This is a formatted string that has a resolution of seconds.
    """
    if not self._is_status_final():
      self._get_job_data()

    if 'start_time' in self._data:
      if '.' in self._data['start_time']:
        return datetime.strptime(self._data['start_time'],
                                 '%Y-%m-%dT%H:%M:%S.%f')
      return datetime.strptime(self._data['start_time'], '%Y-%m-%dT%H:%M:%S')
    else:
      return ''

  @property
  def end_time(self) -> str:
    """
    str: Date and time when the job finished running.

    This is a formatted string that has a resolution of seconds.
    """
    if not self._is_status_final():
      self._get_job_data()

    if 'end_time' in self._data:
      if '.' in self._data['end_time']:
        return datetime.strptime(self._data['end_time'], '%Y-%m-%dT%H:%M:%S.%f')
      elif self._data['end_time'] != '':
        return datetime.strptime(self._data['end_time'], '%Y-%m-%dT%H:%M:%S')
      return ''
    else:
      return ''

  @property
  def default_namespace(self) -> str:
    """
    str: Default namespace of the job.

    .. versionadded:: 2.0.0
    """
    if not self._is_status_final():
      self._get_job_data()

    if 'default_namespace' in self._data:
      return self._data['default_namespace']
    else:
      return None

  @property
  def results_frame(self) -> str:
    """
    str: Name of the results frame for a query job with an INTO clause or None
    for all other job types.

    .. versionadded:: 2.0.1
    """
    if not self._is_status_final():
      self._get_job_data()

    if 'results_frame' in self._data:
      return self._data['results_frame']
    else:
      return None

  @property
  def visited_edges(self) -> Dict[str, int]:
    """
    dict: A dictionary mapping Cypher bound variable names to an integer giving
    the number of edges visited during the job for the Edge Frame referenced by
    the bound variable.

    An edge is "visited" when the query considers the edge as a match to one of
    the query path edges.  Multiple Cypher variables can refer to the same edge
    frame.

    Consider the query path
    ``()-[a:graph_edges1]->()-[b:graph_edges2]->()-[c:graph_edges1]->()`` with
    a visited_edges result of ``a -> 5, b -> 7, c -> 4``.  In performing the
    query 5 edges of type `a` were visited, and so on.  Notice that the total
    number of edges visited for the frame graph_edges1 is 9 while the number of
    edges visited for the frame graph_edges2 is 7.
    """
    if not self._is_status_final():
      self._get_job_data()

    return self._data['visited_edges']

  @property
  def total_visited_edges(self) -> int:
    """
    int: The total number of edges traversed during the job. This is the sum
    of the counts for all edge labels returned in visited_edges.

    For the example given in the visited_edges documentation, the value of
    total_visited_edges would be 16.
    """
    if not self._is_status_final():
      self._get_job_data()

    return sum(self._data['visited_edges'].values())

  @property
  def _timing(self):
    """
    For internal use.
    """
    if not self._is_status_final():
      self._get_job_data()

    return self._data['timing']

  @property
  def description(self) -> str:
    """
    str: A description supplied when the job was started.  Usually a query.
    """
    if not self._is_status_final():
      self._get_job_data()

    return self._data['description']

  @property
  def error_type(self) -> XgtErrorTypes:
    """
    object: Class that belongs to the XgtError hierarchy that corresponds to
    the original exception type thrown that caused the Job to fail.
    """
    if not self._is_status_final():
      self._get_job_data()

    if 'error_type' in self._data:
      return self._data['error_type']
    else:
      return XgtError

  @property
  def error(self) -> str:
    """
    str: User-friendly error message describing the reason a job failed.
    """
    if not self._is_status_final():
      self._get_job_data()

    if 'error' in self._data:
      return self._data['error']
    else:
      return ''

  @property
  def trace(self) -> str:
    """
    str: Very detailed error message for a failed job.

    This error message contains the friendly error message and a stack
    strace for the code that participated in the error.
    """
    if not self._is_status_final():
      self._get_job_data()

    if 'trace' in self._data:
      return self._data['trace']
    else:
      return ''

  @property
  def schema(self) -> List[List]:
    """
    list of lists: The property names and types of the stored results.
    """
    if not self._is_status_final():
      self._get_job_data()

    if 'schema' in self._data:
      return self._data['schema']
    else:
      return None

  @property
  def num_rows(self) -> int:
    """
    int: The number of rows in the query result or the number of correctly
         ingested/inserted rows for an input operation.

    .. versionadded:: 1.15.0
    """
    if not self._is_status_final():
      self._get_job_data()

    if 'num_rows' in self._data:
      return self._data['num_rows']
    else:
      return None

  @property
  def total_ingest_errors(self) -> int:
    """
    int: The number of errors that were thrown during ingest.
    """
    if not self._is_status_final():
      self._get_job_data()

    if 'total_ingest_errors' in self._data:
      return self._data['total_ingest_errors']
    else:
      return None

  def get_data(self, offset : int = 0, length : int = None,
               rows : Iter[int] = None, columns : Iter[Union[int, str]] = None,
               format : str = 'python') -> Union[List[List], pandas.DataFrame,
                                                 pyarrow.Table, None]:

    """
    Returns results data for query jobs with a RETURN but no INTO or None for
    all other job types.

    Parameters
    ----------
    offset : int
      Position (index) of the first row to be retrieved. Cannot be given with
      rows.
      Optional. Default=0.
    length : int
      Maximum number of rows to be retrieved starting from the row
      indicated by offset. A value of 'None' means 'all rows' on and
      after the offset. Cannot be given with rows.
      Optional. Default=None.
    rows : Iterable of int
      The rows to retrieve.  A value of 'None' means all rows.  Cannot be given
      with either offset or length.
      Optional. Default=None.

      .. versionadded:: 1.16.0
    columns : Iterable of int or str
      The columns to retrieve.  Given as an iterable over mixed column
      positions and schema column names.  A value of 'None' means all columns.
      Optional. Default=None.

      .. versionadded:: 1.14.0
    format : str
      Selects the data format returned: a Python list of list, a pandas
      Dataframe, or an Apache Arrow Table.  Must be one of 'python', 'pandas',
      or 'arrow'. Default='python'.

      .. versionadded:: 1.14.0

    Returns
    -------
    Returns one of the following if the job object represents a OpenCypher query with
    no RETURN clause:
      list of lists, pandas DataFrame, or Apache Arrow Table
    Otherwise, returns None.

    Raises
    ------
    ValueError
      If parameter is out of bounds or invalid format given.
    OverflowError
      If data is out of bounds when converting.
    """
    if not 'schema' in self._data:
      return None

    # Get the requested columns as a list of integers.
    schema = self._data['schema']
    columns = _validated_columns(columns, schema)

    # If specific columns are requested, the schema used to convert the pyarrow
    # table to pandas or pandas format should only include those columns.
    if columns is not None:
      schema = []
      for i in columns:
        schema.append(self._data['schema'][i])

    if not isinstance(format, str):
      raise XgtTypeError('format must be of type string')

    # Ignore format case.
    lc_format = format.lower()

    if lc_format == 'python':
      return _get_data_python_from_table(
                 self._load_arrow_table(offset, length, rows, columns),
                 schema, self._conn)
    elif lc_format == 'pandas':
      return _get_data_pandas_from_table(
                 self._load_arrow_table(offset, length, rows, columns),
                 schema, self._conn)
    elif lc_format == 'arrow':
      return self._load_arrow_table(offset, length, rows, columns)
    else:
      raise XgtValueError('format must be one of python, pandas, or arrow, '
                          f'given: {format}')

  def __str__(self):
    txt = f'id:{self.id}, user:{self.user}, description:{self.description}, ' \
          f'start:{self.start_time}, stop:{self.end_time}, status:{self.status}'
    if len(self.error) > 0:
      txt = f'{txt}, nerror:{self.error}'
    return txt

  def get_ingest_errors(self, offset : int = 0,
                        length : int = None) -> List[str]:
    """
    Returns a table of strings giving error information from ingest.  The first
    thousand errors raised are retrievable this way.

    Parameters
    ----------
    offset : int
      Position (index) of the first row to be retrieved.
      Optional. Default=0.
    length : int
      Maximum number of rows of errors to be retrieved starting from the row
      indicated by offset.  A value of 'None' means 'all rows' on and after
      the offset.
      Optional. Default=None.

    Returns
    -------
    list of str
    If this is not an ingest job or no errors were raised, this returns None.
    """
    if not self._is_status_final():
      self._get_job_data()

    if 'ingest_errors' in self._data:
      offset, length = _verify_offset_length(offset, length)

      total_length = len(self._data['ingest_errors'])
      if length is None:
        length = total_length

      errors = self._data['ingest_errors']
      startpos = min(offset, total_length)
      endpos = min(offset + length + 1, total_length)

      return errors[startpos:endpos]
    else:
      return None

  @property
  def _query_plan(self):
    """
    For internal use.
    """
    if not self._is_status_final():
      self._get_job_data()

    if 'query_plan' in self._data:
      return str(self._data['query_plan'])
    else:
      return ''

  @property
  def _query_plan_as_cypher(self):
    """
    For internal use.
    """
    if not self._is_status_final():
      self._get_job_data()

    if 'query_plan' in self._data:
      return self._data['query_plan'].to_cypher()
    else:
      return ''

  @property
  def _query_plan_scores(self):
    """
    For internal use.
    """
    if not self._is_status_final():
      self._get_job_data()

    if 'query_plan' in self._data:
      return self._data['query_plan'].get_scores()
    else:
      return None
